/*
    iprig - ham radio via internet
    Copyright (C) 2002-2006  Ladislav Vaiz <ok1zia@nagano.cz>
    and authors of web browser Links 0.96

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#define LEAK_DEBUG_LISTx
#include "../config.h"

#include <arpa/inet.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#define __USE_GNU
#include <string.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <termios.h>
#include <unistd.h>
#include <stdarg.h>

#ifdef HAVE_LINUX_PPDEV_H
#include <linux/ppdev.h>
#include <linux/parport.h>
#endif

#include "../md5/md5.h"
#include "zstring.h"
#include "audio_descr.h"

#define mem_alloc malloc
#define mem_free free
#define mem_realloc realloc

#ifdef HAVE_ALSA
#define ALSA_PCM_NEW_HW_PARAMS_API
#define ALSA_PCM_NEW_SW_PARAMS_API
#include <alsa/asoundlib.h>
#else
typedef long snd_pcm_sframes_t;
#endif

struct rig; 

static inline char *safe_strncpy(char *dst, const char *src, size_t dst_size) {
    strncpy(dst, src, dst_size);
    if (strlen(src) >= dst_size) dst[dst_size - 1] = 0;
    return dst;
}


static inline char *safe_strncpy0(char *dst, const char *src, size_t dst_size) {
    if (!src) {
        *dst='\0';
        return dst;
    }

    strncpy(dst, src, dst_size);
    if (strlen(src) >= dst_size) dst[dst_size - 1] = 0;
    return dst;
}

/* error.c */
#define internal(args...) debug_internal(__FILE__, __LINE__, args)
/* #define dbg(args...) fprintf(stderr, args)*/
#define error(args...) fprintf(stderr, args)
#define STRERR(no) strerror_r(no, errbuf, sizeof(errbuf))
void dbg(char *m, ...);
void debug_internal(char *file, int line, char *s, ...);
void int_error( char *, ...);

/* main.c */
extern int opt_d, opt_h, opt_i, opt_s, opt_u, opt_v;
extern char *opt_hostname;

void init(void);
void terminate_all_subsystems(void);
char *parse_options(int argc, char **argv);
void usage(void);
void arg0(char *a0);


/* select.c */
#define H_READ  0
#define H_WRITE 1
#define H_ERROR 2
#define H_DATA  3

typedef long ttime;

struct list_head {
    void *next;
    void *prev;
};

struct xlist_head {
    struct xlist_head *next;
    struct xlist_head *prev;
};

#define init_list(x) {(x).next=&(x); (x).prev=&(x);}
#define list_empty(x) ((x).next == &(x))
#define add_at_pos(p,x) do {(x)->next=(p)->next; (x)->prev=(p); (p)->next=(x); (x)->next->prev=(x);} while(0)
#define del_from_list(x) {((struct list_head *)(x)->next)->prev=(x)->prev; ((struct list_head *)(x)->prev)->next=(x)->next;}
//#define add_to_list(l,x) add_at_pos((struct xlist_head *)&(l),(struct xlist_head *)(x))
//#define foreach(e,l) for ((e)=(l).next; (e)!=(void *)&(l); (e)=(e)->next)
#ifdef HAVE_TYPEOF
#define add_to_list(l,x) add_at_pos((typeof(x))&(l),(x))
#define foreach(e,l) for ((e)=(l).next; (e)!=(typeof(e))&(l); (e)=(e)->next)
#define foreachback(e,l) for ((e)=(l).prev; (e)!=(typeof(e))&(l); (e)=(e)->prev)
#else
#define add_to_list(l,x) add_at_pos((struct xlist_head *)&(l),(struct xlist_head *)(x))
#define foreach(e,l) for ((e)=(l).next; (e)!=(void *)&(l); (e)=(e)->next)
#define foreachback(e,l) for ((e)=(l).prev; (e)!=(void *)&(l); (e)=(e)->prev)
#endif

extern int terminate;

void select_loop(void (*init)(void));
#ifdef LEAK_DEBUG_LIST
void debug_set_handlers(char *file, int line, int fd, void (*read_func)(void *), void (*write_func)(void *), void (*error_func)(void *), void * cba);
int debug_install_timer(char *file, int line, ttime t, void (*func)(void *), void * cba);
void debug_kill_timer(char *file, int line, int id);
#else
void set_handlers(int fd, void (*read_func)(void *), void (*write_func)(void *), void (*error_func)(void *), void * cba);
int install_timer(ttime t, void (*func)(void *), void * cba);
void kill_timer(int id);
#endif

#ifdef LEAK_DEBUG_LIST
#define set_handlers(n, r, w, x, a) debug_set_handlers(__FILE__, __LINE__, n, r, w, x, a)
#define install_timer(x, y, z) debug_install_timer(__FILE__, __LINE__, x, y, z)
#define kill_timer(x) debug_kill_timer(__FILE__, __LINE__, x)
#endif

void check_bottom_halves(void);

void normalize_timeval(struct timeval *tv);
void install_signal_handler(int sig, void (*fn)(void *), void * cba, int critical);
    
extern struct timeval start;
#define ST_START gettimeofday(&start, NULL)
#define ST_STOP {\
    struct timeval stop;\
    int sec, usec;\
    gettimeofday(&stop, NULL);\
    usec=stop.tv_usec-start.tv_usec;\
    sec=stop.tv_sec-start.tv_sec;\
    if (usec<0){usec+=1000000;sec--;}\
    dbg("stopky: %s: %d.%06d \n", __FUNCTION__, sec,usec);\
}

/* tcp.c */

#define TCP_PORT 2074
#define MAXRDBUF 1024
#define MAXWRBUF 1024
/* milliseconds */
#define CONNECTION_TIMEOUT 30000
#define SEND_ACK_TIMEOUT 10000
#define ACK_TIMEOUT 20000
#define REMOVE_TIMEOUT 10000

enum tcp_state { 
    TCP_INIT=0, 
    TCP_CONNECTING,
    TCP_CONNECTED,
    TCP_WAIT_ACK,
    TCP_DISCONNECTED,
    TCP_DEAD,  /* conn removed immediately */
};

struct conn {
    struct conn *next;  
    struct conn *prev;
    int sock; 
    enum tcp_state state;
    struct sockaddr_in sin, udpsin;
    int relseq;
    int timer;
    char rdbuf[MAXRDBUF];
    int rdbuflen;
    char wrbuf[MAXWRBUF];
    int wrbuflen;
    int disconnecting;
    char *salt, *user;
    int authorized;
    unsigned long int ssrc;
};

struct conn conns;  	   		/* list of server's connection, valid on server */
extern struct conn *gconn;      /* connection to client with grabbed control */
extern struct conn *sconn; 		/* connection to server, valid only on client */


int init_tcp(void);
void tcp_accept_handler(void *xxx);
void tcp_read_handler(void *xxx);
void tcp_exception_handler(void *xxx);
int tcp_send_state(struct conn *conn);
void tcp_disconnect(struct conn *conn, char *reason);
int conn_prod_state(struct conn *conn);
void tcp_set_state(struct conn *conn, enum tcp_state state);
char *tcp_state_s(enum tcp_state state);
void kill_conn_timer(void *xxx);
void send_ack_timer(void *xxx);
void remove_conn_timer(void *xxx);
void free_conn(struct conn *conn);
void tcp_connect(char *hostname, int port);
void tcp_connected_handler(void *xxx);
void after_connect(struct conn *conn);
void rel_write(struct conn *conn, char *s);
void rel_write_all(char *s);
void rel_read(struct conn *conn, char *line);
void rel_sread(struct conn *conn, char *cmd, zchar *zs);
void rel_cread(struct conn *conn, char *cmd, zchar *zs);
void send_grab(struct conn *conn, int grab);
void client_authorized(struct conn *conn);


/* sdev.c */

/* C5 FCE ADR LEN DATA CHK */

struct sdev{
	int fd;
	char rdbuf[MAXRDBUF];
	int rdbuflen;
	struct conn *conn;
    int timeout_ms;
    char *device;
    void (*callback)(struct rig *rig, char fce, char saddr, char *data, unsigned char len);
};


struct sdev *init_sdev(void);
void free_sdev(struct sdev *sdev);
int sdev_open(struct sdev *sdev);
int sdev_close(struct sdev *sdev);
char sd_chk(char *s, int len);
int sdev_send(struct sdev *sd, char fce, char saddr, char *data, unsigned char len);
int sdev_recv(struct sdev *sd, char *data, unsigned char len);
int sdev_prot(struct sdev *sd, char fce, char saddr, char *data, unsigned char len);
void sdev_read_handler(void *xxx);
void sdev_async(struct sdev *sd, char fce, char saddr, char *data, unsigned char len);
const char *sdev_error(int err);

/* rc.c */

extern struct rcitem cfg;

struct rcitem {
    struct rcitem *next;  
    struct rcitem *prev;
	char *key;
	char *value;
};

int read_rc_file(char *filename);
int read_rc_files(void);
char *rc_getopt(char *key);
void rc_info(void);


/* term.c */

#define KBD_ENTER   0x100
#define KBD_BS      0x101
#define KBD_TAB     0x102
#define KBD_ESC     0x103
#define KBD_LEFT    0x104
#define KBD_RIGHT   0x105
#define KBD_UP      0x106
#define KBD_DOWN    0x107
#define KBD_INS     0x108
#define KBD_DEL     0x109
#define KBD_HOME    0x10a
#define KBD_END     0x10b
#define KBD_PGUP    0x10c
#define KBD_PGDN    0x10d

#define KBD_F1      0x120
#define KBD_F2      0x121
#define KBD_F3      0x122
#define KBD_F4      0x123
#define KBD_F5      0x124
#define KBD_F6      0x125
#define KBD_F7      0x126
#define KBD_F8      0x127
#define KBD_F9      0x128
#define KBD_F10     0x129
#define KBD_F11     0x12a
#define KBD_F12     0x12b

#define KBD_CTRL_C  0x200

#define KBD_SHIFT   1
#define KBD_CTRL    2
#define KBD_ALT     4

#define ESC_TIMEOUT         200

struct term{
  	int fd;
	char rdbuf[MAXRDBUF];
	int rdbuflen;
	struct termios oldtermios;
	int timer;
};


extern struct term *term;

int init_term(void);
void free_term(void);
void term_read_handler(void *xxx);
void term_exception_handler(void *xxx);
int get_esc_code(char *str, int len, char *code, int *num, int *el);
int process_queue(void);
void kbd_timeout(void *);
int handle_key(int key, int shifts);

/* crig.c */

#define MAXGPO 16
#define MAXGPI 16
extern struct crig *crig;

struct crig{
    int grab;
	int ptt, cws, pwr, lna, pa, pwk;
    int gpo[MAXGPO];
};

void init_crig(void);
void free_crig(void);
void crig_grab(void);

/* rig.c */

extern struct rig *rig;

struct rig{
	int ptt, cw, pwr, lna, pa, pwk;
    int gpo[MAXGPO];
    int gpi[MAXGPI];

    void (*rig_free)(struct rig *rig);
    int (*rig_reset)(struct rig *rig);
    /* digital outputs */
    int (*rig_pwr)(struct rig *rig, int on);
    int (*rig_pwk)(struct rig *rig, int on);
    int (*rig_ptt)(struct rig *rig, int on);
    int (*rig_cw)(struct rig *rig, int on);

    int (*rig_lna)(struct rig *rig, int on);
    int (*rig_pa)(struct rig *rig, int on);

    int (*rig_gpo)(struct rig *rig, int gponr, int on);

    /* digital inputs */
    int (*rig_powerok)(struct rig *rig);
    int (*rig_gpi)(struct rig *rig, int gpinr);


    /* ppdev */
    int ppdev_fd;

	/* sior */
	struct sdev *sdev;
};

int init_rig(void);
void free_rig(void);
int rig_reset(void);

/* digital outputs */
int rig_pwr(int on);
int rig_pwk(int on);
int rig_ptt(int on);
int rig_cw(int on);

int rig_lna(int on);
int rig_pa(int on);

int rig_gpo(int gponr, int on);

/* digital inputs */
int rig_powerok(void);
int rig_gpi(int gpinr);   /* reads gpi on request */
void rig_async_gpi(struct rig *rig, int gpinr, int on); /* called from callback by "device driver" */



/* ppdev.c */

int init_ppdev(struct rig *rig);
void free_ppdev(struct rig *rig);
int ppdev_reset(struct rig *rig);

	/* digital outputs */
int ppdev_ptt(struct rig *rig, int on);
int ppdev_cw(struct rig *rig, int on);



/* sior.c */
int init_sior(struct rig *rig);
void free_sior(struct rig *rig);
int sior_reset(struct rig *rig);

	/* digital outputs */
int sior_ptt(struct rig *rig, int on);
int sior_cw(struct rig *rig, int on);
int sior_pwr(struct rig *rig, int on);
int sior_lna(struct rig *rig, int on);
int sior_pa(struct rig *rig, int on);
int sior_gpo(struct rig *rig, int gponr, int on);
int sior_pwk(struct rig *rig, int on);

void sior_async(struct rig *rig, char fce, char saddr, char *data, unsigned char len);
void sior_async_gpi(struct rig *rig, int gpinr, int on);


/* sdump.c */

#define MAXSDUMP 2

struct sdump{
	int fd[MAXSDUMP];
	char rdbuf[MAXRDBUF];
    int colors[MAXSDUMP];
};

struct sdump *init_sdump(void);
int sdump_open(int idx, char *filename);
void sdump_read_handler(void *xxx);

/* mike.c */

#define mike_send(cmd) write(mike_pipe[1], cmd, strlen(cmd))
int mike_set(int compress, int flags, char *nat, int asquelch, int asqdelay);
void mike_rel_read(const char *line);
int mike_run(void);
int mike_kill(void);
int mike_start(void);
int mike_stop(void);

int mike_addest(struct conn *conn);
int mike_rmdest(struct conn *conn);
int mike_getpid(void);
int mike_keepalive(struct conn *conn);
unsigned long int mike_getssrc(void);

/* speaker.c */

#define speaker_send(cmd) write(speaker_pipe[1], cmd, strlen(cmd))
#define speaker_rsend(cmd) write(speaker_rpipe[1], cmd, strlen(cmd))
int speaker_set(int flags, int ajitter, int ajitteridlet);
void speaker_rel_read(const char *line);
int speaker_run(void);
int speaker_kill(void);
int speaker_start(void);
int speaker_stop(void);

int speaker_addest(struct conn *conn);
int speaker_rmdest(struct conn *conn);
int speaker_getpid(void);

void speaker_rread(void *xxx);
void speaker_rrel_read(const char *line);


/* udp.c */

extern int sock;

int init_udp(void);
int free_udp(void);
int udp_keepalive(struct conn *conn);

/* sound.c */
int init_sound(int iomode);
void free_sound(void);
extern void (*sound_term)(int iomode);
extern void (*sound_open)(int *audio_io, int *audio_ctl);
extern void (*sound_play)(int len, unsigned char *buf);
extern void (*sound_playvol)(int value);
extern void (*sound_recgain)(int value);
extern void (*sound_dest)(int where);
extern int  (*sound_grab)(char *buf, int len);
extern void (*sound_flush)(void);
extern snd_pcm_sframes_t (*sound_get_delay)(void);

audio_encoding_t sound_encoding;

/* audio_oss.c */

int oss_init(int iomode);
void oss_term(int iomode);
void oss_open(int *audio_io, int *audio_ctl);
void oss_play(int len, unsigned char *buf);
void oss_playvol(int value);
void oss_recgain(int value);
void oss_dest(int where);
int oss_grab(char *buf, int len);
void oss_flush(void);
snd_pcm_sframes_t oss_get_delay(void);

/* audio_alsa.c */
#ifdef HAVE_ALSA
int alsa_init(int iomode);
void alsa_term(int iomode);
void alsa_open(int *audio_io, int *audio_ctl);
void alsa_play(int len, unsigned char *buf);
void alsa_playvol(int value);
void alsa_recgain(int value);
void alsa_dest(int where);
int alsa_grab(char *buf, int len);
void alsa_flush(void);
snd_pcm_sframes_t alsa_get_delay(void);
#endif

/* zrandom */
int zrandom(void);

/* help */
extern char *help;

