/*
    iprig - ham radio via internet
    Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"
#include "speakfree.h"

int opt_d=1, opt_h=0, opt_i=0, opt_s=0, opt_v=0, opt_u=0;
char *opt_hostname=NULL;
static int ac;
static char **av;

void usage(){
    printf(
"\n"            
" iprig - Remote radio via IP networks\n"
" Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz\n"
"\n"
" Command line arguments:\n"
" -d debug mode\n"
" -h this help plus configuration file format\n"
" -i informations about HW etc\n"
" -s server mode\n"
" -u dump mode (see sdump.c)\n"
" -v version\n"
"\n");
}
         
char *parse_options(int argc, char **argv){
    int c;

    while (1){
        c = getopt(argc, argv, "dhisuv");
        if (c == -1) break;

        switch (c){
            case 'd':
                opt_d++;
                break;

            case 'h':     
                opt_h++;
                break;

            case 'i':
                opt_i++;
                break;

            case 's':
                opt_s++;
                break;
           
            case 'v':
                opt_v++;
                break;

            case 'u':
                opt_u++;
                break;

            default:
                printf ("?? getopt returned character code 0%o ??\n", c);
        }
    }
    if (optind < argc){
        opt_hostname = argv[optind++];
    }

    if (optind < argc){
        printf ("non-option ARGV-elements: ");
        while (optind < argc) printf ("%s ", argv[optind++]);
        printf ("\n");
    }

    return NULL;
}
     

void sigchld(void *xxx){
    int pid, status;
    char *name;
    
    
    
    while ((pid = waitpid(0, &status, WNOHANG)) > 0){
        name="";
        if (pid==mike_getpid()) name="mike";
        if (pid==speaker_getpid()) name="speaker";
        
        error("Child %d (%s) exited, status=%d ", pid, name, status); 
        if (WIFEXITED(status)) error("normally=%d ", WEXITSTATUS(status));
        if (WIFSIGNALED(status)) error("signal=%d ", WTERMSIG(status));
        if (WIFSTOPPED(status)) error("stopped=%d ", WSTOPSIG(status));
        dbg("\n");
    }

}

void init(){
    int flags=0, jitter=100000, jitteridlet=400000;
    char *c;
    
    parse_options(ac, av);
    
    if (opt_v){
        printf ("iprig 0.1\nCopyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>\n");
        exit(0);
    }
    
    if (opt_h){
        usage();
		printf("\nConfiguration file /etc/iprig or ~/.iprig has format:\n"
				"key value\n\n"
				"Keys are:\n"
				"%s", help);
        exit(0);
    }

    read_rc_files();
    if (opt_i){
        rc_info();
        exit(0);
    }
    
    if (opt_u){
        init_sdump();
        return;
    }
    
    if (opt_d){
        dbg("debug level %d\n", opt_d);
    }

	setpgrp();
    srand(time(NULL));
    install_signal_handler(SIGCHLD, sigchld, NULL, 0);
    
    c = rc_getopt("codec"); /* Audio codec for mike (alaw mulaw gsm ilbc) */ 
    if (c && strcmp(c, "mulaw")==0){
        dbg("codec: mulaw\n");
        sound_encoding = AE_PCMU;
    }else if (c && strcmp(c, "gsm")==0){
        dbg("codec: gsm, format=mulaw\n");
        sound_encoding = AE_PCMU; 
        flags|=fCompGSM;
    }else if (c && strcmp(c, "ilbc")==0){
        dbg("codec: ilbc, format=mulaw\n");
        sound_encoding = AE_PCMU; 
        flags|=fCompILBC;
    }else{
        /* A-law is default */
        dbg("codec: alaw\n");
        sound_encoding = AE_PCMA;
    }
    

    c = rc_getopt("jitter"); /* Jitter compensation delay in us (100000) */
    if (c) jitter=atoi(c); 
    
    c = rc_getopt("jitteridle"); /* Countdown to reset jitter timer in us (100000) */
    if (c) jitteridlet=atoi(c); 
    
    
    if (opt_s){
        /* server */
        init_rig();
        init_tcp();
        init_udp();
        mike_set(flags, 0, NULL, 0, 12000);
        speaker_set(0, jitter, jitteridlet);
        mike_run();
        speaker_run();
    }else{
        /* client */
		init_crig();
        init_udp();
        mike_set(flags, 0, NULL, 0, 12000);
        speaker_set(0, jitter, jitteridlet);
        mike_run();
        speaker_run();
        
        if (!opt_hostname){
            opt_hostname = rc_getopt("server"); /* Hostname/IP of server to connect to */
            if (!opt_hostname){
                usage();
                exit(0);
            }
        }
		init_term();
        tcp_connect(opt_hostname, 2074);
    }
}

void terminate_all_subsystems(){
	dbg("\ncleanup\n");
	mike_kill();
	speaker_kill();
    free_udp();
	if (!opt_s){
		free_term();
	}
}

int main(int argc, char *argv[])
{
    ac = argc;
    av = (char **)argv;

    
    select_loop(init);
    terminate_all_subsystems();
    
    return 0;
}

void arg0(char *a0){
    int i;
    for (i=0; i<ac; i++) memset(av[i], 0, strlen(av[i]));
    strcpy(av[0], a0);
}
