/*
    iprig - ham radio via internet
    Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

/* #define PGP_DEBUG */
/*

                        Speak Freely for Unix
                  Network sound transmission program

        Designed and implemented in July of 1991 by John Walker

*/

#include "speakfree.h"
//#include "version.h"
#include "header.h"
#include "zstring.h"

#ifdef HAVE_DEV_RANDOM

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#endif

static struct destination *dests = NULL/*, *dtail*/;
static int gsmcompress = FALSE;       /* GSM compress buffers */
static int ilbccompress = FALSE;      /* ILBC compress buffers */

#define DEFAULT_SQUELCH 4096          /* Default squelch level */
static int squelch = 0;               /* Squelch level if > 0 */
static int sqdelay = 12000;           /* Samples to delay before squelch */
static int sqwait = 0;                /* Squelch delay countdown */
static int push = TRUE;               /* Push to talk mode */
static int talking = FALSE;           /* Push to talk button state */
static int agc = FALSE;               /* Automatic gain control active ? */
static int rgain = 33;                /* Current recording gain level */
static int spurt = TRUE;              /* True for first packet of talk spurt */
static int debugging = FALSE;         /* Debugging enabled here and there ? */
static int havesound = FALSE;         /* True if we've acquired sound input */
static char hostname[20];             /* Host name to send with packets */
//static int loopback = FALSE;          /* Remote loopback mode */
static gsm gsmh;                      /* GSM handle */
static iLBC_Enc_Inst_t Enc_Inst;
static iLBC_Dec_Inst_t Dec_Inst;
//static struct adpcm_state adpcm = {0, 0}; /* ADPCM compression state */


static unsigned long ssrc;            /* RTP synchronisation source identifier */
static unsigned long timestamp;       /* RTP packet timestamp */
static unsigned short seq;            /* RTP packet sequence number */
static unsigned long rtpdesrand;      /* RTP DES random RTCP prefix */

static char *sdes = NULL;             /* RTP SDES packet */
static int sdesl;                     /* RTP SDES packet length */

static int sound_packet;              /* Ideal samples/packet */

//#define TimerStep   (7 * 1000000L)    /* Alarm interval in microseconds */

#ifdef HEWLETT_PACKARD
extern int HPTermHookChar;
#endif

static int mike_pid = 0;
static int mike_pipe[2];
static int mike_running = 1;

#define ucase(x)    (islower(x) ? toupper(x) : (x))

#ifndef USE_CURSES

#ifndef UNIX420
#define UNIX5
#endif

#ifdef UNIX420
#include <sgtty.h>
#endif /* UNIX420 */


#endif
/*  ULARM  --  Wrapper for setitimer() that looks like alarm()
               but accepts a time in microseconds.  */

/*static void ularm(long t)
{
    struct itimerval it;

    it.it_value.tv_sec = t / 1000000L;
    it.it_value.tv_usec = t % 1000000L;
    it.it_interval.tv_sec = it.it_interval.tv_usec = 0;
    setitimer(ITIMER_REAL, &it, NULL);
} */

/*  GSMCOMP  --  Compress the contents of a sound buffer using GSM.  */

static void gsmcomp(struct soundbuf *sb)
{

    gsm_frame dst;
    int i, l = 0;
    short int slen;
    char *dp = ((char *) sb->buffer.buffer_val) + sizeof(short);
    gsm_signal src[160];
    int j;

    //ST_STOP;
    //dbg("gsmcomp() len(mlaw)=%d", sb->buffer.buffer_len);
    sb->compression |= fCompGSM;
    for (i = 0; i < sb->buffer.buffer_len; i += 160) {
        for (j = 0; j < 160; j++) {
            if ((i + j) < sb->buffer.buffer_len) {
                src[j] = audio_u2s(sb->buffer.buffer_val[i + j]);
            } else {
                src[j] = 0;
            }
        }
        gsm_encode(gsmh, src, dst);
        bcopy(dst, dp, sizeof dst);
        dp += sizeof dst;
        l += sizeof dst;
    }
    slen=sb->buffer.buffer_len;
    //dbg(" l(gsm)=%d\n", l);
    //ST_START;
    /* Hide original uncompressed buffer length in first 2 bytes of buffer. */
    
    *((short *) sb->buffer.buffer_val) = htons(slen);
    sb->buffer.buffer_len = l + sizeof(short);
    //dbg("gsmcomp: slen=%d buffer_len=%d\n", slen, sb->buffer.buffer_len);
}

static void ilbccomp(struct soundbuf *sb)
{
    unsigned char dst[100];
    int i, l = 0;
    char *dp = (char *) sb->buffer.buffer_val;
    float block[BLOCKL_MAX];
    int j;

    //dbg("ilbccomp: buffer_len=%d\n", sb->buffer.buffer_len);
    sb->compression |= fCompILBC;
    for (i = 0; i < sb->buffer.buffer_len; i += 160) {
        for (j = 0; j < 160; j++) {
            if ((i + j) < sb->buffer.buffer_len) {
                block[j] = (float)audio_u2s(sb->buffer.buffer_val[i + j]);
            } else {
                block[j] = 0.0;
            }
        }
       /* {
            int i;
            for (i=0;i<20; i++) dbg("%02x ", (unsigned char) block[i]);
            dbg("\n");
        }*/ 
        iLBC_encode(dst, block, &Enc_Inst);
        /*{
            int i;
            for (i=0;i<38; i++) dbg("%02x ", (unsigned char) dst[i]);
            dbg("\n");
        } 
        iLBC_decode(block, dst, &Dec_Inst, 1); 
        {
            int i;
            for (i=0;i<20; i++) dbg("%02x ", (unsigned char) block[i]);
            dbg("\n");
        } */
        bcopy(dst, dp, Enc_Inst.no_of_bytes);
        dp += Enc_Inst.no_of_bytes;
        l += Enc_Inst.no_of_bytes;
    }
    sb->buffer.buffer_len = l;
    //dbg("ilbccomp: output len=%d\n", l);
}

#ifdef USE_RTCP
/*  SENDRTPCTRL  --  Send RTP  control message, encrypting if
                     necessary.  */

static int sendrtpctrl(struct destination *d, struct sockaddr *destaddr,
                       char *msg, int msgl)
{
//	dbg("sending control packet\n");
    return sendto(d->sock, msg, msgl, 0, destaddr, sizeof(struct sockaddr));
}
#endif

void dumpsin(char *text, struct sockaddr_in *sin){
    long naddr;
	
	naddr = ntohl(sin->sin_addr.s_addr);
    dbg("%s %d.%d.%d.%d:%d\n", text, (naddr>>24)&0xff, (naddr>>16)&0xff, (naddr>>8)&0xff, naddr&0xff, ntohs(sin->sin_port));
	
}

static int addest(const char *line)
{
    struct destination *d;
    zchar *zs;
    char *cmd, *arg;
    
  //  dbg("mike: addest: %s\n", line);

    zs = zstrdup(line);
    cmd = ztokenize(zs, 1);

    d = (struct destination *) malloc(sizeof(struct destination));
	memset(d, 0, sizeof (struct destination));

	arg = ztokenize(zs, 0);
    d->sin.sin_family = AF_INET;
    inet_aton(arg, &d->sin.sin_addr);
	arg = ztokenize(zs, 0);
    d->sin.sin_port=htons(atoi(arg));

	arg = ztokenize(zs, 0);
    d->tcpsock = atoi(arg);
	
	arg = ztokenize(zs, 0);
    d->ssrc = strtoul(arg, NULL, 16);
    
    zfree(zs);
    d->dnext = dests;
    dests = d;
    d->sock = sock;

    return TRUE;
}

static int rmdest(const char *line)
{
    struct destination *d, *prev;
    zchar *zs;
    char *cmd, *arg;
    int tcpsock;
#ifdef USE_RTCP    
    char v[1024];
    int l;
#endif    
    
   // dbg("mike: rmdest: %s\n", line);

    zs = zstrdup(line);
    cmd = ztokenize(zs, 1);

	arg = ztokenize(zs, 0);
    tcpsock = atoi(arg);
    zfree(zs);

    for (d = prev = dests; d != NULL; prev = d, d = d->dnext) {
        if (d->tcpsock != tcpsock) continue;
        dbg("found socket %d\n", tcpsock);
        if (d==dests) 
            dests = d->dnext;
        else
            prev->dnext = d->dnext;
#ifdef USE_RTCP
        l = rtp_make_bye((unsigned char *) v, ssrc, "rmdest", TRUE);
		dbg("sending BYE1 to %d (%d) ssrc=%08x\n", d->tcpsock, d->sock, ssrc);
        if (sendrtpctrl(d, (struct sockaddr *) &(d->sin), v, l) < 0) {
            perror("sending RTCP BYE packet");
        }
#endif        
        
        free(d);
    }
    return TRUE;
}

/*  TIMERTICK  --  Timer tick signal-catching function.  */

#if 0
static void timertick(int sig)
{
    struct destination *d;

    if (talking){
/*        for (d = dests; d != NULL; d = d->dnext) {
            if (sendrtpctrl(d, (struct sockaddr *) &(d->sin), sdes, sdesl) < 0) {
                perror("sending RTCP SDES packet");
            }
        }  */
    }

    ularm(TimerStep);
    signal(SIGALRM, timertick);       /* Reset signal to handle timeout */
}
#endif

/*  SENDPKT  --  Send a message to all active destinations.  */

static int sendpkt(struct soundbuf *sb)
{
    struct destination *d;
    int pktlen;
	static char errbuf[256];

//    dbg("sendpkt(buffer_len=%db)\n", sb->buffer.buffer_len);
    //dbg("sendpkt gsmcompress=%d ilbccompress=%d\n", gsmcompress, ilbccompress);
    if (gsmcompress) {
        gsmcomp(sb);
    }else if (ilbccompress){
        ilbccomp(sb);
    }else{
        if (sound_encoding==AE_PCMA) 
            sb->compression|=fEncAlaw;
        else
            sb->compression&=~fEncAlaw;
    }
    
    pktlen = sb->buffer.buffer_len + (sizeof(struct soundbuf) - BUFL);
   // dbg("sendpkt: buffer_len=%db, pktlen=%db\n", sb->buffer.buffer_len, pktlen); 

//    dbg("sendpkt sb->compression=%d spurt=%d\n", sb->compression, spurt);
    pktlen = rtpout(sb, ssrc, timestamp, seq, spurt);
    seq++;
    timestamp += sound_packet;
    //dbg("sendto(rtp) len=%d\n", pktlen);

    spurt = FALSE;                    /* Not the start of a talk spurt */

    for (d = dests; d != NULL; d = d->dnext) {
        if (sendto(d->sock, (char *) sb, pktlen,
            0, (struct sockaddr *) &(d->sin), sizeof d->sin) < 0) {

            error("sendto %s:%d %s", inet_ntoa(d->sin.sin_addr), ntohs(d->sin.sin_port), STRERR(errno));
            return FALSE;
        }
    }
    return TRUE;
}

/*  OPENAUDIO  --  Open audio input.  If audio hardware is half duplex,
                  we may have to mute output in progress.  */

static int openaudio(void)
{
    int i;

    i = init_sound(O_RDONLY);

    if (i <= 0) {
        fprintf(stderr, "iprig: unable to initialise audio.\n");
        return FALSE;
    }
    havesound = TRUE;
    if (agc) {
        sound_recgain(rgain);      /* Set initial record level */
    }
    return TRUE;
}

/*  FREEAUDIO  --  Release sound input device, sending a resume to
                   the output device if we've muted it due to
                   half-duplex hardware.  */

static void freeaudio(void)
{
    if (havesound) {
        if (debugging) {
            fprintf(stderr, "Restoring audio modes at release.\n");
        }
        sound_term(O_RDONLY);
        havesound = FALSE;
    }

}

/*  EXITING  --  Catch as many program termination signals as
                 possible and restore initial audio modes before
                 we exit.  */

static void exiting(int sig)
{
#ifdef USE_RTCP
    struct destination *d;
#endif

    freeaudio();

#ifdef USE_RTCP
    for (d = dests; d != NULL; d = d->dnext) {
        char v[1024];
        int l;

        l = rtp_make_bye((unsigned char *) v, ssrc, "Exiting iprig", TRUE);
		dbg("sending BYE2 to %d (%d) ssrc=%08x\n", d->tcpsock, d->sock, ssrc);
        if (sendrtpctrl(d, (struct sockaddr *) &(d->sin), v, l) < 0) {
            perror("sending RTCP BYE packet");
        }
    }          
#endif    
    dbg("mike exiting (sig %d)\n", sig);
    exit(0);
}

/*  CHANGEMODE  --  Reset state when talk/pause/quiet mode changes.  */

static void changemode(void)
{
    if (talking) {
        spurt = TRUE;
        /* Discard all backlog sound input. */
        sound_flush();
    } else {
        sqwait = 0;                   /* Reset squelch delay */
    }
}


void mike_rel_read(const char *line){
    zchar *zs;
    char *cmd, *arg1/*, *c*/;
#ifdef USE_RTCP    
    char v[1024];
    int l;
#else    
    int pktlen;
#endif
    int tcpsock;
    struct destination *d;
    struct soundbuf sb;
    char errbuf[300];

    if (strlen(line)==0) {
        error("rel_read: got empty line\n");
        return;
    }

    zs = zstrdup(line);
    cmd = ztokenize(zs, 1);

	if (strcmp(cmd, "KA")==0){
		arg1 = ztokenize(zs, 0);
		tcpsock=atoi(arg1);
        for (d = dests; d != NULL; d = d->dnext) {
        	if (d->tcpsock != tcpsock) continue;
#ifdef USE_RTCP    
			//dbg("sending SDES to %d (%d) ssrc=%08x\n", d->tcpsock, d->sock, ssrc);
            if (sendrtpctrl(d, (struct sockaddr *) &(d->sin), sdes, sdesl) < 0) {
                perror("sending RTCP SDES packet");
            }
#else            
            sb.compression = fKeepAlive;
            pktlen = rtpout(&sb, ssrc, timestamp, seq, 0);
            if (sendto(d->sock, (char *) &sb, pktlen, 0, (struct sockaddr *) &(d->sin), sizeof d->sin) < 0) {
                error("sendto KA %s:%d %s", inet_ntoa(d->sin.sin_addr), ntohs(d->sin.sin_port), STRERR(errno));
            }
#endif            
        } 
		goto x;
	}
    dbg("mike_rel_read('%s')\n", line);
    if (strcmp(cmd, "q")==0){   
        mike_running=0;
        goto x;
    }
    if (strcmp(cmd, "0")==0){   
        //dbg("mike stop\n");
        if (!talking) goto x;
        talking= FALSE;
        changemode();
        
#ifdef USE_RTCP        
        for (d = dests; d != NULL; d = d->dnext) {
            l = rtp_make_bye((unsigned char *) v, ssrc, "rx", TRUE);
			dbg("sending BYE3 to %d (%d) ssrc=%08x\n", d->tcpsock, d->sock, ssrc);
            if (sendrtpctrl(d, (struct sockaddr *) &(d->sin), v, l) < 0) {
                perror("sending RTCP BYE packet");
            }
        }
#endif
        goto x;
    }
    if (strcmp(cmd, "1")==0){   
        //dbg("mike start\n");
        if (talking) goto x;
        talking=TRUE;	
        changemode();
        goto x;
    }
    if (strcmp(cmd, "+")==0){
        //dbg("cmd +\n");
        addest(line);
        goto x;
    }
    if (strcmp(cmd, "-")==0){
        //dbg("cmd -\n");
        rmdest(line);
        goto x;
    }
    
x:;
    zfree(zs);
}


/*  SENDFILE  --  Send a file or, if the file name is NULL or a
                  single period, send real-time sound input. */

static int sendfile(char *f)
{ 
    soundbuf netbuf;
//#define buf netbuf.buffer.buffer_val
    int /*bread, nread,*/ lread;
//    FILE *afile = NULL;
    fd_set fdset;
    char errbuf[256];
    char s[1024], *d;
    char rdbuf[MAXRDBUF];
    int ret, len, rdbuflen;
    struct timeval tv;

	signal(SIGHUP, exiting);     /* Set signal to handle termination */
	signal(SIGINT, exiting);     /* Set signal to handle termination */
	signal(SIGTERM, exiting);    /* Set signal to handle termination */

    /* Compute the number of sound samples needed to fill a
       packet of TINY_PACKETS bytes. */

    if (gsmcompress){
        sound_packet = 160 * 4;
    }else if (ilbccompress){
        sound_packet = Enc_Inst.blockl*4;
    }else{
        sound_packet = 320;
    }    
 //   sound_packet = (gsmcompress) ? (160 * 4): 320;

#define SHOW_PACKET_SIZE
#ifdef SHOW_PACKET_SIZE
    printf("Samples per packet = %d\n", sound_packet);
#endif
    lread = sound_packet;

    strcpy(netbuf.sendinghost, hostname);

	{
        fcntl(fileno(stdin), F_SETFL, O_NDELAY);
        //talking = TRUE;
        spurt = TRUE;

        /* Send real-time sound. */

#ifdef Solaris
        setaubufsize(sound_packet);
#endif

        if (!openaudio()) {
            fprintf(stderr, "Unable to initialise audio.\n");
            return 2;
        }
        changemode();
        mike_running=1;
        rdbuflen=0;
        while (mike_running) {
            int soundel = 0;
            unsigned char *bs = (unsigned char *) netbuf.buffer.buffer_val;

            //dbg("talking=%d  havesound=%d \n", talking, havesound);
            if (havesound) {
                /* Obtain a packet's worth of real-time audio from
                   the audio input hardware. */
                soundel = sound_grab(netbuf.buffer.buffer_val, sound_packet);
             //   memset(netbuf.buffer.buffer_val, 0xff, soundel);
                //dbg("\nsound_grab(%db)=%db\n", sound_packet, soundel); 
                //dbg("soundel=%d\n", soundel);
           /*     {
                    int i;
                    for (i=0;i<20; i++) dbg("%02x ", (unsigned char) netbuf.buffer.buffer_val[i]);
                    dbg("\n");
                } */
            }

        
            FD_ZERO(&fdset);
            FD_SET(0, &fdset);
            tv.tv_sec=0;
            tv.tv_usec=0;
            if (select(1, &fdset, NULL, NULL, &tv) < 0) {
                if (errno==EINTR) continue;
                error("mike: select failed %s\n", STRERR(errno));
                exiting(0);
            }

            if (FD_ISSET(0, &fdset)){
                ret = read(0, s, sizeof(s));
                if (ret<0) {
                    error("mike: read(0) failed\n");/* maybe SIGPIPE */
                    exiting(0);
                }
                if (ret<0) {
                    if (errno==EINTR) continue;
                    error("mike: read(0) return 0\n");/* maybe SIGPIPE */
                    exiting(0);
                }
                if (rdbuflen + ret >= MAXRDBUF){
                    error("mike: rdbuf overflow\n");
                    exiting(0);
                }
                memcpy(rdbuf + rdbuflen, s, ret);
                rdbuflen += ret;

                while(1){
                    d = index(rdbuf, '\n');
                    if (!d) break;
                    len = d - rdbuf;
                    rdbuf[len] = '\0';
                    if (len>0) {
                        if (rdbuf[len - 1] == '\r') rdbuf[len - 1] = '\0';
                        mike_rel_read(rdbuf);
                    }
                    memmove(rdbuf, rdbuf + len + 1, rdbuflen - len - 1);
                    rdbuflen -= len + 1;
                }
            }
            //dbg("talking=%d  soundel=%d\n", talking, soundel);
            if (!talking) continue;
            if (soundel > 0) {
                int squelched = (squelch > 0);
                register unsigned char *start = bs;
                register int j;

                /* If entire buffer is less than squelch, ditch it.  If
                   we haven't received sqdelay samples since the last
                   squelch event, continue to transmit. */

                if (sqdelay > 0 && sqwait > 0) {
                    if (debugging) {
                        printf("Squelch countdown: %d samples left.\n",
                            sqwait);
                    }
                    sqwait -= soundel;
                    squelched = FALSE;
                    if (sqwait <= 0 && push) {
                        fprintf(stderr, "mike: Quiet (2)\n");
                    }
                } else if (squelch > 0) {
                    for (j = 0; j < soundel; j++) {
#ifdef USQUELCH
                        if (((*start++ & 0x7F) ^ 0x7F) > squelch)
#else
                        int samp = audio_u2s(*start++);

                        if (samp < 0) {
                            samp = -samp;
                        }
                        if (samp > squelch)
#endif
                        {
                            squelched = FALSE;
                            if (sqwait <= 0 && push) {
                               fprintf(stderr, "\rTalk:  ");
                            }
                            sqwait = sqdelay;
                            break;
                        }
                    }
                }
                if (squelched) {
                    if (debugging) {
                        printf("Entire buffer squelched.\n");
                    }
                    spurt = TRUE;
                } else {
                    netbuf.compression = 0;
                    netbuf.compression |= debugging ? fDebug : 0;
//                    netbuf.compression |= loopback ? fLoopBack : 0;

                    /* If automatic gain control is enabled,
                       ride the gain pot to fill the dynamic range
                       optimally. */
                    if (agc) {
                        register unsigned char *start = bs;
                        register int j;
                        long msamp = 0;

                        for (j = 0; j < soundel; j++) {
                            int s = audio_u2s(*start++);

                            msamp += (s < 0) ? -s : s;
                        }
                        msamp /= soundel;
                        if (msamp < 6000) {
                            if (rgain < 100) {
                                sound_recgain(++rgain);
                            }
                        } else if (msamp > 7000) {
                            if (rgain > 1) {
                                sound_recgain(--rgain);
                            }
                        }
                    }

                    netbuf.buffer.buffer_len = soundel;
                    if (!sendpkt(&netbuf)) {
                        exiting(-1);
                    }
                }
            } else {
                usleep(100000L);  /* Wait for some sound to arrive */
            }
        }
    }
    return 0;
}

/*  MAKESESSIONKEY  --  Generate session key with optional start
                        key.  If mode is TRUE, the key will be
                        translated to a string, otherwise it is
                        returned as 16 binary bytes.  */

static void makeSessionKey(char *key, char *seed, int mode)
{
    int j, k;
    struct MD5Context md5c;
    unsigned char md5key[16], md5key1[16];
#define SESSIONKEYBUF 1024
    char s[SESSIONKEYBUF];
#ifdef HAVE_DEV_RANDOM
    int random_fd, wanted_random_bytes, random_bytes;
    char rand[SESSIONKEYBUF];
#endif

    s[0] = 0;
    if (seed != NULL) {
        strcat(s, seed);
    }

    /* The following creates a seed for the session key generator
       based on a collection of volatile and environment-specific
       information unlikely to be vulnerable (as a whole) to an
       exhaustive search attack.  If one of these items isn't
       available on your machine, replace it with something
       equivalent or, if you like, just delete it. */

    sprintf(s + strlen(s), "%lu", (unsigned long) getpid());
    sprintf(s + strlen(s), "%lu", (unsigned long) getppid());
    V getcwd(s + strlen(s), 256);
    sprintf(s + strlen(s), "%lu", (unsigned long) clock());
    sprintf(s + strlen(s), "%lu", (unsigned long) time(NULL));
#ifdef Solaris
    sysinfo(SI_HW_SERIAL,s + strlen(s), 12);
#else
    sprintf(s + strlen(s), "%lu", (unsigned long) gethostid());
#endif
    getdomainname(s + strlen(s), 256);
    gethostname(s + strlen(s), 256);
    sprintf(s + strlen(s), "%u", getuid());
    sprintf(s + strlen(s), "%u", getgid());

#ifdef HAVE_DEV_RANDOM
    /* Some systems a random device which offers data as long as the system
       has still enough "randomness" collected.  We use it here. */
 
    /* Leave room for the terminator and to append data for second hash. */
    wanted_random_bytes = SESSIONKEYBUF - (strlen(s) - 12);

    if (wanted_random_bytes > 1) {
        dbg("opening /dev/random\n");
        random_fd = open ("/dev/random",O_RDONLY,O_NONBLOCK);
        if (random_fd > 0) { /* Do not complain if open fails */
            random_bytes = read(random_fd, rand, wanted_random_bytes);

            if  (random_bytes  > 0) { /* Again do not complain on read errors */

                if (random_bytes <= wanted_random_bytes) { /* Should be */
                    /* String termination will be lost after memcpy */
                    char *startcopy =  s + strlen(s); 
                    memcpy(startcopy, rand, random_bytes);
                    startcopy[random_bytes + 1] = 0; /* Terminate string */
                }
            }
            close(random_fd);
        }
        dbg("closed /dev/random\n");
    }
#endif

    MD5Init(&md5c);
    MD5Update(&md5c, (unsigned char *)s, strlen(s));
    MD5Final(md5key, &md5c);
    /*  If you append additional information here, verify that the
        computation of wanted_random_bytes above leaves adequate room after
        the /dev/random data for its maximum length.  */
    sprintf(s + strlen(s), "%lu", (unsigned long) ((time(NULL) + 65121) ^ 0x375F));
    MD5Init(&md5c);
    MD5Update(&md5c, (unsigned char *)s, strlen(s));
    MD5Final(md5key1, &md5c);
#define nextrand    (md5key[j] ^ md5key1[j])
    if (mode) {
        for (j = k = 0; j < 16; j++) {
            unsigned char rb = nextrand;

/* #define Rad16(x) ((x) + 'A') */
    /* We don't count on choosing where in the alphabet to
       start based on the hashes to stir in any more entropy
       than the fundamental 4 bits from the generated key
       (although it may), but primarily to spread out the
       letter distribution across the entire alphabet rather
       than just the first 16 letters. */
#define Rad16(x)  ((((((md5key1[j] >> 1) ^ (md5key[15 - j] >> 2)) % 26) + (x)) % 26) + 'A')
            key[k++] = Rad16((rb >> 4) & 0xF);
            key[k++] = Rad16(rb & 0xF);
#undef Rad16
            if (j & 1) {
                key[k++] = '-';
            }
        }
        key[--k] = 0;
    } else {
        for (j = 0; j < 16; j++) {
            key[j] = nextrand;
        }
    }
#undef nextrand
}


/*  Main program.  */


int mike_set(int compress, int flags, char *nat, int asquelch, int asqdelay){

    gsmcompress = FALSE;
    ilbccompress = FALSE;
    if (compress & fCompGSM) gsmcompress = TRUE;
    if (compress & fCompILBC) ilbccompress = TRUE;
    
    if (flags & FLG_DEBUG) debugging = TRUE;
    if (flags & FLG_AGC) agc = TRUE;
    squelch = asquelch;
    sqdelay = asqdelay;

   	return 0; 
}
int mike_run()
{
    char md5key[16];
/*	int i;*/

	if (mike_pid) {
		dbg("mike already running\n");
        return 1;
	}
    
	/* Initialise randomised packet counters. */

    makeSessionKey(md5key, NULL, FALSE);
    bcopy(md5key, (char *) &ssrc, sizeof ssrc);
    bcopy(md5key + sizeof ssrc, (char *) &timestamp,
          sizeof timestamp);
    bcopy(md5key + sizeof ssrc + sizeof timestamp,
          (char *) &seq, sizeof seq);
    bcopy(md5key + sizeof ssrc + sizeof timestamp +
          sizeof seq, &rtpdesrand, sizeof rtpdesrand);
    sdesl = rtp_make_sdes(&sdes, ssrc, -1, TRUE);
	dbg("rtp_make_sdes ssrc=%08x\n",ssrc);

	if (pipe(mike_pipe)){
		error("Can't create mike pipe\n");
		return -1;
	}
	
	mike_pid=fork();
	if (mike_pid<0){  /* error */
		error("Can't fork mike\n");
		return -1;
	}
	if (mike_pid>0) { /* parent */
		close(mike_pipe[0]);
		return 0;
	}
	/* child */
	close(mike_pipe[1]);
	close(0);
	if (dup(mike_pipe[0])){
		error("Can't dup\n");
		return -1;
	}
	close(mike_pipe[0]);
/*	for (i=3;i<16;i++) close(i);*/

    arg0("iprig mike");
    dbg("mike started (server=%d)\n", opt_s);
	
	
    if (gethostname(hostname, sizeof hostname) == -1) {
        strcpy(hostname, "localhost");
    } else {
        if (strlen(hostname) > 15) {
            hostname[15] = 0;
        }
    }
    
    gsmh = gsm_create();
    initEncode(&Enc_Inst, 20);
    initDecode(&Dec_Inst, 20, 1);


    return sendfile(NULL);

    exiting(-1);
    gsm_destroy(gsmh);
    return 0;
}


int mike_kill(void){
    if (!mike_pid) {
        error("mike_start: mike_pid==0\n");
        return 0;
    }

	mike_send("q\n");
    kill(mike_pid, SIGINT);
	dbg("waiting for mike pid %d\n", mike_pid);
    ST_START;
	waitpid(mike_pid, NULL, 0);
    ST_STOP;
	//dbg("mike pid %d terminated\n", mike_pid);
	mike_pid = 0;
	
	return(0);
}

int mike_start(void){
    if (!mike_pid) {
        error("mike_start: mike_pid==0\n");
        return 0;
    }
	mike_send("1\n");

	return(0);
}

int mike_stop(void){
    if (!mike_pid) {
        error("mike_stop: mike_pid==0\n");
        return 0;
    }

	mike_send("0\n");
	return(0);
}

int mike_addest(struct conn *conn){
    char *c;
    char s1[20], s2[20], s3[20], s4[20];

    strcpy(s1, inet_ntoa(conn->udpsin.sin_addr));
    sprintf(s2, "%d", ntohs(conn->udpsin.sin_port));
    sprintf(s3, "%d", conn->sock); 
    sprintf(s4, "0x%08lx", conn->ssrc); 
    c=zconcatesc("+", s1, s2, s3, s4, NULL);
    mike_send(c);
    dbg("mike_addest(%s)\n", c);
    mike_send("\n");
    mem_free(c);
    return 0;
}

int mike_rmdest(struct conn *conn){
    char *c;
    char s1[20];

    sprintf(s1, "%d", conn->sock); 
    c=zconcatesc("-", s1, NULL);
    mike_send(c);
    dbg("mike_rmdest(%s)\n", c);
    mike_send("\n");
    mem_free(c);
    return 0;
}

int mike_getpid(){
    return mike_pid;
}

int mike_keepalive(struct conn *conn){
    char *c;
    char s1[20];

    sprintf(s1, "%d", conn->sock); 
    c=zconcatesc("KA", s1, NULL);
    mike_send(c);
    mike_send("\n");
	return 0;
}

unsigned long int mike_getssrc(void){
	return ssrc;
}
