/*
    iprig - ham radio via internet
    Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

int init_ppdev(struct rig *rig){
	int mode;
	struct ppdev_frob_struct frob;
    char errbuf[1024];
    char *device;
    
    dbg("init_ppdev\n");

    rig->rig_free   = free_ppdev;
    rig->rig_ptt    = ppdev_ptt;
    rig->rig_cw     = ppdev_cw;

    device = rc_getopt("ppdev_device"); /* Path to ppdev device (/dev/parport0) */
    rig->ppdev_fd = open(device, O_RDWR|O_NONBLOCK);
    if (rig->ppdev_fd < 0){
        error("Can't open %s %s", device, strerror_r(errno, errbuf, sizeof(errbuf)));
        close(rig->ppdev_fd);
        rig->ppdev_fd = -1;
        return -1;
    }

    mode=PARPORT_MODE_PCSPP;
    if (ioctl(rig->ppdev_fd, PPSETMODE, &mode)<0){
        dbg("Can't set %s to EPP mode %s", device, STRERR(errno));
        close(rig->ppdev_fd);
        rig->ppdev_fd=-1;
        return -1;
    }

    if (ioctl(rig->ppdev_fd, PPEXCL, NULL)<0){
        dbg("Can't exclusive access to %s %s", device, STRERR(errno));
        close(rig->ppdev_fd);
        rig->ppdev_fd=-1;
        return -1;
    }

    if (ioctl(rig->ppdev_fd, PPCLAIM, NULL)<0){
        dbg("Can't claim %s %s", device, STRERR(errno));
        close(rig->ppdev_fd);
        rig->ppdev_fd=-1;
        return -1;
    }
    
	frob.mask = PARPORT_CONTROL_STROBE;
	frob.val = 0;
	if (ioctl(rig->ppdev_fd, PPFCONTROL, &frob)<0){
        dbg("Can't clear parport_STROBE %s", STRERR(errno)); 
        close(rig->ppdev_fd);
        rig->ppdev_fd=-1;
        return 1;
    }
    
    rig_reset();

	frob.mask = PARPORT_CONTROL_STROBE;
	frob.val = PARPORT_CONTROL_STROBE;
	if (ioctl(rig->ppdev_fd, PPFCONTROL, &frob)<0){
        dbg("Can't set parport_STROBE %s", STRERR(errno)); 
        close(rig->ppdev_fd);
        rig->ppdev_fd=-1;
        return 1;
    }

    return 0;
}

void free_ppdev(struct rig *rig){
	struct ppdev_frob_struct frob;
    char errbuf[1024];
    
    if (!rig) return;
	
    if (rig->ppdev_fd>=0){
        frob.mask = PARPORT_CONTROL_STROBE;
        frob.val = 0;
        if (ioctl(rig->ppdev_fd, PPFCONTROL, &frob)<0){
            error("Can't clear parport_STROBE %s\n", strerror_r(errno, errbuf, sizeof(errbuf))); 
        }
        close(rig->ppdev_fd);
        rig->ppdev_fd=-1;
    }
	return;
}

/* digital outputs */
int ppdev_ptt(struct rig *rig, int on){
    struct ppdev_frob_struct frob;
    char errbuf[1024];
    
	frob.mask = PARPORT_CONTROL_INIT;
	frob.val = on?PARPORT_CONTROL_INIT:0;
	if (ioctl(rig->ppdev_fd, PPFCONTROL, &frob)<0){
        dbg("Can't %s parport_INIT %s\n", on?"set":"clear", strerror_r(errno, errbuf, sizeof(errbuf))); 
        close(rig->ppdev_fd);
        rig->ppdev_fd=-1;
        return 1;
    }
    return 0; 
}

int ppdev_cw(struct rig *rig, int on){
	struct ppdev_frob_struct frob;
    char errbuf[1024];
    
	frob.mask = PARPORT_CONTROL_SELECT;
	frob.val = !on?PARPORT_CONTROL_SELECT:0;
	if (ioctl(rig->ppdev_fd, PPFCONTROL, &frob)<0){
        dbg("Can't %s parport_SELECT %s\n", on?"set":"clear", strerror_r(errno, errbuf, sizeof(errbuf))); 
        close(rig->ppdev_fd);
        rig->ppdev_fd=-1;
        return 1;
    }
    return 0;

}

