/*
    iprig - ham radio via internet
    Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

struct rig *rig = NULL;

int init_rig(){
    int ret;
    char *rigtype;

   // dbg("init_rig\n");
    rigtype = rc_getopt("rig_type"); /* Type of rig (ppdev, sior, siop) */
    if (!rigtype){
        error("rig_type unspecified\n");
        return -1;
    }
	rig = mem_alloc(sizeof(struct rig));
	memset(rig, 0, sizeof(struct rig));

    if (strcasecmp(rigtype, "ppdev")==0) {
        ret = init_ppdev(rig);
        if (ret) return ret;
    }
    if (strcasecmp(rigtype, "sior")==0) {
        ret = init_sior(rig);
        if (ret) return ret;
    }
    return 0;
}

void free_rig(){
    if (rig && rig->rig_free) rig->rig_free(rig);
	mem_free(rig);
    rig = NULL;
}

int rig_reset(void){
    if (!rig) return -1;

    rig_ptt(0);
    rig_cw(0);
    
    if (!rig->rig_reset) return 0;
    return rig->rig_reset(rig);
}

/* digital outputs */
int rig_pwr(int on){
    if (!rig || !rig->rig_pwr) return -1;
    return rig->rig_pwr(rig, on);
}

int rig_pwk(int on){
    if (!rig || !rig->rig_pwk) return -1;
    return rig->rig_pwk(rig, on);
}

int rig_ptt(int on){
    if (!rig || !rig->rig_ptt) return -1;
    return rig->rig_ptt(rig, on);
}

int rig_cw(int on){
    if (!rig || !rig->rig_cw) return -1;
    return rig->rig_cw(rig, on);
}

int rig_lna(int on){
    if (!rig || !rig->rig_lna) return -1;
    return rig->rig_lna(rig, on);
}

int rig_pa(int on){
    if (!rig || !rig->rig_pa) return -1;
    return rig->rig_pa(rig, on);
}

int rig_gpo(int gponr, int on){
    if (!rig || !rig->rig_gpo) return -1;
    return rig->rig_gpo(rig, gponr, on);
}

/* digital inputs */
int rig_powerok(){
    if (!rig || !rig->rig_powerok) return -1;
    return rig->rig_powerok(rig);
}

int rig_gpi(int gpinr){
    if (!rig || !rig->rig_gpi) return -1;
    return rig->rig_gpi(rig, gpinr);
}

void rig_async_gpi(struct rig *rig, int gpinr, int on){ /* called from callback by "device driver" */
    switch(gpinr){
        case 0: /* GPI0 */
            break;
        case 10: /* +8V */
            break;
    }
    return;
}
