/*
    iprig - Remote radio via IP
    Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

//#define STARTB 4
//#define STOPB 1


struct sdev *init_sdev(){
    struct sdev *sdev;

	sdev=mem_alloc(sizeof(struct sdev));
	bzero(sdev, sizeof(struct sdev));
	sdev->fd=-1;
	sdev->rdbuflen=0;
	sdev->conn=NULL;

	return sdev;
}

void free_sdev(struct sdev *sdev){
    if (sdev->fd >= 0) sdev_close(sdev);
    if (sdev->device) free(sdev->device);
    free(sdev);   
}

int sdev_open(struct sdev *sdev){
	char errbuf[1030];
	struct termios tio;
    int i;
	
	sdev->rdbuflen=0;
	sdev->conn=NULL;
    sdev->timeout_ms=10;

	sdev->fd=open(sdev->device, O_RDWR | O_SYNC | O_NONBLOCK | O_NOCTTY);
	if (sdev->fd < 0){
		error("error opening file %s - %s\n", sdev->device, STRERR(errno));
		return -1;
	}
	//dbg("sdev opened, fd=%d\n", sdev->fd);
    
	if (fcntl(sdev->fd, F_SETFL, O_NONBLOCK)) {
		error("error setting nonblock %s\n", STRERR(errno));
		sdev_close(sdev);
        return -1;
    }
    
	if (tcgetattr(sdev->fd,&tio)){
		error("tcgetattr failed %s\n", STRERR(errno));
		sdev_close(sdev);
        return -1;
	}
	tio.c_cflag=B115200|CS8|CLOCAL|CREAD/*|PARENB*/;/*|PARODD;*/
	tio.c_iflag=INPCK;
	tio.c_lflag=0;
	tio.c_oflag=0;
	tio.c_cc[VMIN]=1;
	tio.c_cc[VTIME]=5;
    
	if (tcsetattr(sdev->fd,TCSANOW,&tio)){
		error("tcsetattr failed %s\n", STRERR(errno));
		sdev_close(sdev);
        return -1;
	}

#if 1    
    i = TIOCM_RTS; /* clear PSEN*/
    ioctl (sdev->fd, TIOCMBIC, &i); /* clear = log.1*/
    i = TIOCM_DTR; /* clear reset*/
    ioctl (sdev->fd, TIOCMBIC, &i); /* clear */
    usleep(100000);
#endif

    set_handlers(sdev->fd, sdev_read_handler, NULL, NULL, sdev);

    
	return sdev->fd;
}

int sdev_close(struct sdev *sdev){
	struct termios tio;

	if (sdev->fd<0) return 0;

    set_handlers(sdev->fd, NULL, NULL, NULL, NULL);

	sdev->rdbuflen=0;
	sdev->conn=NULL;

	if (tcgetattr(sdev->fd,&tio)!=0){
        tio.c_cflag=B115200|CS8|CLOCAL|CREAD;
        tio.c_iflag=INPCK;
        tio.c_lflag=0;
        tio.c_oflag=0;
        tio.c_cc[VMIN]=1;
        tio.c_cc[VTIME]=5;
        
        tcsetattr(sdev->fd,TCSANOW,&tio);
	}
    
	close(sdev->fd);
	sdev->fd=-1;
	dbg("sdev %d closed\n", sdev->fd);
	return 0;
}

char sdev_chk(char *s, int len){
	char chk, *c;
    chk=0;
	for (c=s;len;c++,len--){
		chk^=*c;
	}
	return chk;
}

int sdev_send(struct sdev *sd, char fce, char saddr, char *data, unsigned char len){
	char rawdata[300];
	int rawlen,written;

	rawlen=0;
#ifdef STARTB    
	memset(rawdata, 0xff, STARTB); rawlen+=STARTB;
#endif    
	rawdata[rawlen++]=0xc5;
	rawdata[rawlen++]=fce&0x7f;
	rawdata[rawlen++]=saddr;
	rawdata[rawlen++]=len;
    memcpy(rawdata+rawlen, data, len); rawlen+=len;
#ifdef STARTB    
	rawdata[rawlen]=sdev_chk(rawdata+STARTB, rawlen-STARTB);
#else    
	rawdata[rawlen]=sdev_chk(rawdata, rawlen);
#endif
    rawlen++;
#ifdef STOPB    
	memset(rawdata+rawlen, 0xff, STOPB); rawlen+=STOPB;
#endif    
	
	written=write(sd->fd, rawdata, rawlen);
   /* dbg("sdev_send: written=%d\n", written);*/
    if (written<0) return written;

#if 0
    int i;
    dbg("sdev_send(%d): ", rawlen);
    for (i=0;i<rawlen;i++) dbg("%02x ", (unsigned char)rawdata[i]);
    dbg("\n");
#endif    
    
    return 0;
}


int sdev_recv(struct sdev *sd, char *data, unsigned char len){
    struct timeval tv;
    fd_set rd;
    int ret, i;

    
    tv.tv_usec = len * 1000000/(115200/10) + sd->timeout_ms * 1000;
    tv.tv_sec = 0;
    normalize_timeval(&tv);
    /*dbg("tv=%d.%06d\n", tv.tv_sec, tv.tv_usec);*/
     
    FD_ZERO(&rd);
    FD_SET(sd->fd, &rd);
    ret=select(sd->fd + 1, &rd, NULL, NULL, &tv);
    if (ret<0) return ret;
    if (ret==0) return -13;
    if (!FD_ISSET(sd->fd, &rd)) return -2;
    
    ret=read(sd->fd, data, len);
    if (ret==0) dbg("read() returns 0\n");  
#if 1
    dbg("sdev_recv(%d): ", len);
    for (i=0;i<ret;i++) dbg("%02x ", (unsigned char)data[i]);
    dbg("\n");
#endif    
    return ret;
}

/* C5 FCE ADR LEN DATA CHK */

int sdev_prot(struct sdev *sd, char fce, char saddr, char *data, unsigned char len){
    int ret;
    char rawdata[300];


    /* clearing queue, filedescriptor is non-blocking */
    while (read(sd->fd, rawdata, sizeof(rawdata))>0){};
    /*if (tcflush(sd->fd, TCIFLUSH)){
        dbg("error calling tcflush(%d)\n", sd->fd);
    } */
    
    ret=sdev_send(sd, fce, saddr, data, len);
    if (ret<0) return ret;
/*    if (tcdrain(sd->fd)){
        dbg("error calling tcdrain(%d)\n", sd->fd);
    } */

    while(1) {
        ret=sdev_recv(sd, rawdata, 1);
        if (ret<0) return ret;
        if (rawdata[0]!=(char)0xc5) continue;
        break;
    }
    
    ret=sdev_recv(sd, rawdata+1, 3);
    if (ret<0) return ret;
    
    ret=sdev_recv(sd, rawdata+4, rawdata[3]+1);
    if (ret<0) return ret;

    if (sdev_chk(rawdata, 5+rawdata[3])){
        return -11;
    }
        
    /* correct checksum */
    if (rawdata[1]==0){       /* unknown function */
        return -17;
    }
    if (rawdata[1]==(char)128){     /* error */
        return -14;
    }
    if ((rawdata[1]&0x7f) != (fce&0x7f)){   /* bad answer */
        return -16;
    }
    return 0;
}

void sdev_read_handler(void *xxx){
    struct sdev *sdev;
    int ret, max, i, lastc5;

    sdev=(struct sdev *)xxx;
    max=MAXRDBUF-sdev->rdbuflen;
    if (max<0) internal("sdev_read_hanler: max<0");

    ret=read(sdev->fd, sdev->rdbuf + sdev->rdbuflen, max);
    if (ret<0) internal("sdev_read_handler: read returns <0");
#if 1
    dbg("sdev_read_handler(%d): ", ret);
    for (i=0;i<ret;i++) dbg("%02x ", (unsigned char)sdev->rdbuf[sdev->rdbuflen+i]);
    dbg("\n");
#endif    
    sdev->rdbuflen += ret;

    lastc5=0;
    for (i=0; i < sdev->rdbuflen; i++){
        if (sdev->rdbuf[i] != '\xc5') continue;
        if (i + 5 > sdev->rdbuflen) break; /* to short packet */
        if (i + 5 + sdev->rdbuf[i+3] > sdev->rdbuflen) break;

        if (sdev_chk(sdev->rdbuf+i, 5 + sdev->rdbuf[i+3])){
            continue; /* -11 - Bad checksum*/
        }
        if (sdev->rdbuf[i+1] == '\0'){
            continue; /* -17 - Unknown function*/
        }
        if (sdev->rdbuf[i+1] == '\x80'){
            continue; /* -14 - Not completed */
        }
        if ((sdev->rdbuf[i+1]&0x80)==0x00){  
            /* command, ignoring */
            continue;
        }
        dbg("i=%d\n", i);
        lastc5=i;
        if (sdev->callback) 
            sdev->callback(rig, sdev->rdbuf[i+1], sdev->rdbuf[i+2], sdev->rdbuf + i + 4, sdev->rdbuf[i+3]);
        else 
            sdev_async(sdev, sdev->rdbuf[i+1], sdev->rdbuf[i+2], sdev->rdbuf + i + 4, sdev->rdbuf[i+3]);
        i += 5 + sdev->rdbuf[3] - 1;  /* skip processed data */
    }

    for (i=lastc5; i < sdev->rdbuflen; i++){
        if (sdev->rdbuf[i] != '\xc5') continue;
        if (i==0) break;
        memmove(sdev->rdbuf, sdev->rdbuf+i, sdev->rdbuflen - i);
        sdev->rdbuflen -= i;
        break;
    }

}

void sdev_async(struct sdev *sd, char fce, char saddr, char *data, unsigned char len){
    int i;

    dbg("sdev_async(, %d(%d), %d) ", (unsigned char)fce, fce&0x7f, (unsigned char)saddr);
    for (i=0; i<len; i++) dbg("%02x ", (unsigned char)data[i]);
    dbg("\n");
}

const char *sdev_error(int err){
    static char errstr[100];
    switch(err){
        case  0: return "No error";
        case -11: return "Bad checksum";
        case -13: return "Timeout";
        case -14: return "Not completed";
        case -16: return "Bad answer";
        case -17: return "Unknown function";
    }
    sprintf(errstr, "Unknown error code %d", err);
    return errstr;
}
