/*
    iprig - ham radio via internet
    Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

int init_sior(struct rig *rig){
//    char errbuf[1024];
    char *device;
    
    dbg("init_sior\n");

    rig->rig_free   = free_sior;
    rig->rig_reset  = sior_reset;
    rig->rig_ptt    = sior_ptt;
    rig->rig_cw     = sior_cw;
    rig->rig_pwr    = sior_pwr;
    rig->rig_lna    = sior_lna;
    rig->rig_pa     = sior_pa;
    rig->rig_gpo    = sior_gpo;
    rig->rig_pwk    = sior_pwk;


    rig->sdev = init_sdev();
    device = rc_getopt("sior_device"); /* Path to sior device (/dev/ttyS0) */
    if (!device) device="/dev/ttyS0";
    rig->sdev->device=strdup(device);
    rig->sdev->callback=sior_async;
    sdev_open(rig->sdev);
    
    rig_reset();

    return 0;
}

void free_sior(struct rig *rig){
//    char errbuf[1024];
    
    if (!rig) return;
	
    if (rig->sdev){
        sdev_close(rig->sdev);
        free_sdev(rig->sdev);
        rig->sdev=NULL;
    }
	return;
}

int sior_reset(struct rig *rig){
    int ret;

    ret=sdev_prot(rig->sdev, 64, 240, "", 0);
	if (ret<0){
        dbg("Can't reset sior: %s\n", sdev_error(ret)); 
        return 1;
    }

    /* turn on inputs IO0, +8V and INT */
    ret=sdev_prot(rig->sdev, 66, 240, "\x01\x0c", 2);
	if (ret<0){
        dbg("Can't set active inputs on sior: %s\n", sdev_error(ret)); 
        return 1;
    }
    return 0; 
}

/* digital outputs */
int sior_ptt(struct rig *rig, int on){
    int ret;

    ret=sdev_prot(rig->sdev, 65, 240, on?"\x02\x00":"\x02\x01", 2);
	if (ret<0){
        dbg("Can't %s sior_PTT_2: %s\n", on?"set":"clear", sdev_error(ret)); 
        return 1;
    }
    return 0; 
}

int sior_cw(struct rig *rig, int on){
    int ret;

    ret=sdev_prot(rig->sdev, 65, 240, on?"\x01\x00":"\x01\x01", 2);
	if (ret<0){
        dbg("Can't %s sior_CW_1: %s\n", on?"set":"clear", sdev_error(ret)); 
        return 1;
    }
    return 0;
}

int sior_pwr(struct rig *rig, int on){
    int ret;

    ret=sdev_prot(rig->sdev, 65, 240, on?"\x03\x00":"\x03\x01", 2);
	if (ret<0){
        dbg("Can't %s sior_PWR_3: %s\n", on?"set":"clear", sdev_error(ret)); 
        return 1;
    }
    return 0;
}

int sior_lna(struct rig *rig, int on){
    int ret;

    ret=sdev_prot(rig->sdev, 65, 240, on?"\x04\x00":"\x04\x01", 2);
	if (ret<0){
        dbg("Can't %s sior_LNA_4: %s\n", on?"set":"clear", sdev_error(ret)); 
        return 1;
    }
    return 0;
}

int sior_pa(struct rig *rig, int on){
    int ret;

    ret=sdev_prot(rig->sdev, 65, 240, on?"\x05\x00":"\x05\x01", 2);
	if (ret<0){
        dbg("Can't %s sior_PA_5: %s\n", on?"set":"clear", sdev_error(ret)); 
        return 1;
    }
    return 0;
}

int sior_gpo(struct rig *rig, int gponr, int on){
    char s[10];
    int ret;

    s[0]=gponr;
    s[1]=on?'\x00':'\x01';
    ret=sdev_prot(rig->sdev, 65, 240, s, 2);
	if (ret<0){
        dbg("Can't %s sior_GPO%d: %s\n", on?"set":"clear", gponr, sdev_error(ret)); 
        return 1;
    }
    return 0;
}

int sior_pwk(struct rig *rig, int on){
    int ret;

    ret=sdev_prot(rig->sdev, 65, 240, on?"\x0f\x00":"\x0f\x01", 2);
	if (ret<0){
        dbg("Can't %s sior_PWK: %s\n", on?"set":"clear", sdev_error(ret)); 
        return 1;
    }
    return 0;
}

void sior_async(struct rig *rig, char fce, char saddr, char *data, unsigned char len){
    int i, j;

    dbg("sdev_async(, %d(%d), %d) ", (unsigned char)fce, fce&0x7f, (unsigned char)saddr);
    for (i=0; i<len; i++) dbg("%02x ", (unsigned char)data[i]);
    dbg("\n");
    /* io0, int, 8v */

    switch(fce&0x7f){
        case 64:
        case 67:
            for (j=0; j<2; j++){
                for (i=0; i<8; i++){
                    if (((data[j]&(1<<i))!=0) == rig->gpi[i+j*8]) continue;
                    sior_async_gpi(rig, i+j*8, (data[j]&(1<<i))!=0);
                    rig->gpi[i+j*8] = (data[j]&(1<<i))!=0; 
                }
            }
            break;
    }
}

void sior_async_gpi(struct rig *rig, int gpinr, int on){ 
    switch(gpinr){
        case 0: /* GPI0 */
        case 10: /* +8V */
            rig_async_gpi(rig, gpinr, on);
            break;
        case 11: /* INT */
            if (on) break; /* rising edge */
            /* TODO enumerate I2C devices, clear INT */
            break;
    }
}
