/*
    iprig - ham radio via internet
    Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "speakfree.h"
#include "header.h"

void (*sound_term)(int iomode);
void (*sound_open)(int *audio_io, int *audio_ctl);
void (*sound_play)(int len, unsigned char *buf);
void (*sound_playvol)(int value);
void (*sound_recgain)(int value);
void (*sound_dest)(int where);
int  (*sound_grab)(char *buf, int len);
void (*sound_flush)(void);
snd_pcm_sframes_t (*sound_get_delay)(void);

audio_encoding_t sound_encoding;

int init_sound(int iomode){
    char *drv;
    int ret;

    drv=rc_getopt("sound_type"); /* Sound type (oss, alsa) */

    if (drv && strcmp(drv, "alsa")==0){
#ifdef HAVE_ALSA        
        sound_term = alsa_term;
        sound_open = alsa_open;
        sound_play = alsa_play;
        sound_playvol = alsa_playvol;
        sound_recgain = alsa_recgain;
        sound_dest = alsa_dest;
        sound_grab = alsa_grab;
        sound_flush = alsa_flush;
        sound_get_delay = alsa_get_delay;
        
        ret = alsa_init(iomode);
#else
        error("Alsa not supported");
#endif        
    }else{
        sound_term = oss_term;
        sound_open = oss_open;
        sound_play = oss_play;
        sound_playvol = oss_playvol;
        sound_recgain = oss_recgain;
        sound_dest = oss_dest;
        sound_grab = oss_grab;
        sound_flush = oss_flush;
        sound_get_delay = oss_get_delay;
        
        ret = oss_init(iomode);
    }
    switch(sound_encoding){
        case AE_PCMU:
            break;
        case AE_PCMA:
            break;
        default:
            break;
    }
    
    return ret;
}

void free_sound(){
    sound_term(0);
    sound_term(O_RDONLY);
}

snd_pcm_sframes_t oss_get_delay(void){
    return 0;
}
