/*
    iprig - ham radio via internet
    Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

/* #define PGP_DEBUG */
/*

                        Speak Freely for Unix
                         Sound Output Server

        Designed and implemented in July of 1991 by John Walker.

*/

#include "speakfree.h"
//#include "version.h"
#include "header.h"

#define DEBUG_JITTERx

static int audiok = FALSE;            /* Audio initialised flag */
static int debugging = FALSE;         /* Debugging enabled */
static int debugforce = FALSE;        /* Debugging forced / prevented ? */
#ifdef HEXDUMP
static int hexdump = FALSE;           /* Dump received packets in hex ? */
#endif
//static int noring = FALSE;            /* Disable remote ring requests */
gsm gsmh;                             /* GSM handle */
iLBC_Dec_Inst_t Dec_Inst;
/*static int sock;*/                      /* Data socket */
/*static int ssock;*/                     /* Control socket (RTP/VAT) */
static struct sockaddr_in from;       /* Sending host address */
//static struct sockaddr_in name;       /* Address of destination host */
static socklen_t fromlen;                   /* Length of sending host address */
static int rll;                       /* Length of packet read from socket */
static int Cjitter = 0;               /* Jitter compensation delay */
static int jitterwait = 0;            /* Jitter wait in progress */
static int jitteridle = 0;            /* Countdown to reset jitter timer */
static int Cjitteridlet = 0;          /* Jitter idle timeout interval */

struct jitterbuf {                    /* Jitter queue item */
    struct jitterbuf *next;
    u_int16 seq;                      /* sequence number */
    LONG length;
    char value[2];
};
static struct jitterbuf *jithead = NULL, *jittail = NULL; /* Jitter queue links */

//static int adjustRate = FALSE;        /* Dynamically adjust output rate ? */
static long lastPacketLength = 0;     /* Samples received since last packet */
#define DYNRATE_UPDATE_SAMPLE_COUNT SAMPLE_RATE /* Update dynamic rate after this many samples received */
#define DYNRATE_SMOOTHING_FACTOR 0.1  /* Exponential moving average smoothing factor for rate */

/* Audio output and control device file names. */

#define Debug   ((debugforce != 0) ? (debugforce > 0) : debugging)

#define TickTock    (10 * 1000000L)   /* Alarm interval in microseconds */
#define TockTock    (60 * 1000000L)   /* Alarm interval when no connections open */
#define JitterTock  (1000000L / 4)    /* Alarm interval for jitter compensation */

static int timerStep = TockTock;      /* Current timer step */

static int speaker_pid = 0;
static int speaker_pipe[2];
static int speaker_rpipe[2];
static int speaker_running = 1;
static struct destination *dests = NULL/*, *dtail*/;

static void sigalrm(int sig);

/*  GSMDECOMP  --  Uncompress the contents of a sound buffer using GSM.  */

static void gsmdecomp(struct soundbuf *sb)
{
    int i, j, l = 0;
    char *dpx = ((char *) sb->buffer.buffer_val) + sizeof(short);
    static char dcb[BUFL];
    short declen = ntohs(*((short *) sb->buffer.buffer_val));
    gsm_signal dst[160];

    /* Since some old releases forget to convert this field
       from network byte order, kludge it back to sanity
       if this happens.  First we try swapping the bytes.  If
       that doesn't work, force the standard buffer length.
       Note that we have to write this out, as ntohs will be
       a no-op on a big-endian machine. */

    if (declen <= 0 || declen > 1600) {
        declen = ((declen & 0xFF) << 8) | ((declen >> 8) & 0xFF);
    }
    if (declen <= 0 || declen > 1600) {
        declen = 1600;
    }
    for (i = 0; i < sb->buffer.buffer_len - sizeof(short);
                i += sizeof(gsm_frame)) {
        gsm_decode(gsmh, (gsm_byte *) dpx, dst);
        dpx += sizeof(gsm_frame);
        for (j = 0; j < 160; j++) {
            dcb[l++] = audio_s2u(dst[j]);
        }
    }
    bcopy(dcb, sb->buffer.buffer_val, declen);
   // dbg("gsmdecomp: declen=%d buffer_len=%d\n", declen, sb->buffer.buffer_len);
    sb->buffer.buffer_len = declen;
}

static void ilbcdecomp(struct soundbuf *sb){
    int i, j, l = 0;
    char *dpx = (char *) sb->buffer.buffer_val;
    static char dcb[BUFL];
    float dst[100];
    int block;

    if (sb->buffer.buffer_len%38==0){ 
        block=38;
    }else{
        block=50;
    }
//    dbg("ilbcdecomp: buffer_len=%d mode=%d\n", sb->buffer.buffer_len, mode);

    for (i = 0; i < sb->buffer.buffer_len; i += block) {
        /*{
            int i;
            for (i=0;i<38; i++) dbg("%02x ", (unsigned char) dpx[i]);
            dbg("\n");
        } */
        iLBC_decode(dst, (unsigned char*)dpx, &Dec_Inst, 1); 
        /*{
            int i;
            for (i=0;i<20; i++) dbg("%02x ", (unsigned char) dst[i]);
            dbg("\n");
        } */
     //   dbg("  %d\n", Dec_Inst.blockl);
        dpx += block;
        for (j = 0; j < Dec_Inst.blockl; j++) {
            dcb[l++] = audio_s2u((short)dst[j]);
        }
    }
    bcopy(dcb, sb->buffer.buffer_val, l);
//    dbg("ilbcdecomp: l=%d buffer_len=%d\n", l, sb->buffer.buffer_len);
    sb->buffer.buffer_len = l;
//    dbg("ilbcdecomp: l=%d buffer_len=%d\n", l, sb->buffer.buffer_len);

}


/*  COMPRESSIONTYPE  --  Return a string describing the type of
                         compression employed in this buffer.  Assumes
                         this *is*, in fact, a sound buffer and not
                         an RTP packet, face data, etc.  */

static char *compressionType(soundbuf *msg)
{
    int compression;

    compression = ntohl(msg->compression);
    return ((compression & fCompGSM) ==
                           fCompGSM) ?
                           "GSM compressed" :
           ((compression & fCompGSM) ? "GSM compressed" :
                                       "uncompressed");
}


/*  ULARM  --  Wrapper for setitimer() that looks like alarm()
               but accepts a time in microseconds.  */

static void ularm(long t)
{
    struct itimerval it;

    it.it_value.tv_sec = t / 1000000L;
    it.it_value.tv_usec = t % 1000000L;
    it.it_interval.tv_sec = it.it_interval.tv_usec = 0;
    setitimer(ITIMER_REAL, &it, NULL);
}

/*  WINDTIMER  --  Reset alarm timer to appropriate value
                   depending on what's going on at the moment.  */

static void windtimer(void)
{
    if (audiok) {
        timerStep = (jitteridle > 0) ? JitterTock : TickTock;
    } else {
        timerStep = TockTock;
    }
    ularm(timerStep);
}

/*  OPENAUDIO  --  Attempt to obtain the audio output device.  */

static void openaudio(void)
{
    if (!audiok) {
    //    dbg("speaker: openaudio\n");
        if (!init_sound(O_WRONLY)) {
            perror("opening audio output device");
            return;                   /* Can't acquire sound device */
        }
        audiok = TRUE;
        if (Debug) {
            fprintf(stderr, "speaker: opening audio device.\n");
        }
        windtimer();                  /* Reset timer, if necessary */
    }
}

static void freeaudio(void){
    //dbg("speaker: freeaudio\n");
    if (!audiok) return;
    sound_term(0);
    audiok = FALSE;
}

/*  FLUSHJITTER  --  Flush any buffers waiting in the jitter compensation
                     queue.  */

static void flushjitter(int sig)
{
    struct jitterbuf *jb;

    /* If any samples are pending in the jitter queue, flush
       them to the audio output.  */

    if ((jb = jithead) == NULL) return;
    jithead = jittail = NULL;

    if (Debug) {
        fprintf(stderr, "Flushing jitter queue.\n");
    }
    openaudio();
    while (jb != NULL) {
        struct jitterbuf *jo = jb;

        if (Debug) {
            fprintf(stderr,
                    "Playing %ld samples from jitter queue.\n", jb->length);
        }
        if (audiok) {
            struct timeval start, stop;
            int sec, usec;
            /*{
                int i;
                for (i=0;i<20; i++) dbg("%02x ", (unsigned char)jb->value[i]);
                dbg("\n");
            } */
#ifdef DEBUG_JITTER            
            dbg("playing enqueued %05d delay=%d\n", jb->seq, sound_get_delay());
#endif
            gettimeofday(&start, NULL);
            sound_play(jb->length, (unsigned char *) jb->value);
            gettimeofday(&stop, NULL);
            usec=stop.tv_usec-start.tv_usec;
            sec=stop.tv_sec-start.tv_sec;
            if (usec<0){usec+=1000000;sec--;}
            if (sec>0 || (sec==0 && usec>100000)){
                dbg("sound_play blocked for %d.%06d sec\n", sec, usec);
            }
        }
        jb = jb->next;
        free(jo);
    }
}

/*  SIGALRM  --  Signal-catching function which releases the audio
                 device if we haven't received anything to play in
         the last minute. */

static void sigalrm(int sig)
{
    /* If sound is waiting in the jitter queue, decrement the
       jitter timer and flush the queue if it decrements to zero. */
#ifdef DEBUG_JITTER            
    dbg("*** SIGALRM *** \n");
#endif

    if (jithead != NULL && jitterwait > 0) {
        jitterwait -= timerStep;
        if (jitterwait <= 0) {
            jitterwait = 0;
            flushjitter(1);
        }
    }

    /* Count down the jitter idle timer.  When it's zero, the
       first packets of an audio stream will be buffered in
       the jitter queue. */

    if (jitteridle > 0) {
        jitteridle -= timerStep;
        if (jitteridle <= 0) {
            jitteridle = 0;
            if (Debug) {
                fprintf(stderr, "speaker: jitter idle\n");
            }
        }
    }

    /* If we still own the sound device or have a connection
       open, reset the timer. */

    windtimer();
    if (Debug) {
        time_t t = time(NULL);
        fprintf(stderr, "speaker: Tick: %.2f... %s",
        timerStep / 1000000.0, ctime(&t));
    }
    signal(SIGALRM, sigalrm);         /* Set signal to handle timeout */
}

/*  EXITING  --  Catch as many program termination signals as
                 possible and clean up before exit.  */

static void exiting(int sig)
{

/*    if (sdes) {
        sendLwlMessage(TRUE);
    }*/

    freeaudio();
    dbg("speaker exiting\n");
    exit(0);
}


/*  PLAYBUFFER  --  Send next buffer to audio output. */

static void playbuffer(soundbuf *msg, u_int16 seq)
{
    char *val;
    LONG len;
    unsigned long cmodes;

    debugging = (msg->compression & fDebug) ? TRUE : FALSE;
    
    openaudio();                    /* Open audio output if necessary */

    len = msg->buffer.buffer_len;
    val = msg->buffer.buffer_val;
    
    /*  Long we get here, we should have already verified that the
        length in the packet is consistent with the length of
        data actually read from the socket which, in turn, is
        guaranteed to be limited to BUFL--our maximum buffer size.
        Just as a sanity check, and prior to the specific checks
        made below based on details of the protocol, make sure
        something didn't slip by which might cause us to index
        off the end of a buffer.  We allow for worst-case round-up
        of the size in the buffer due to encryption in making
        the comparison.  */
        
    if (((len + 15) & (~15)) >= BUFL) {
        fprintf(stderr, "speaker: ** Sound buffer in %s of %ld bytes\n"
                        "       exceeds maximum size of %d bytes--discarded. **\n",
            compressionType(msg), len, BUFL);
        return;
    }

    if (Debug) {
        fprintf(stderr, "speaker: playing %ld %s bytes.\n", len,
                compressionType(msg));
    }


    /* With the exception of 2X compression, which may be selected
       by itself or combined with any other mode, compression modes
       are mutually exclusive.  Verify that this is the case in the
       packet before proceeding with decompression.  Any possible
       damage which may be caused by this will be averted by the
       length checks in the decoding sequence below, but identifying
       the source of the error before we blither into the
       decoders makes diagnosis easier, which is important since
       packets specifying multiple compression modes are an obvious
       way malicious users may try to provoke buffer overflows. */
       
    cmodes = msg->compression & fCompressionModes;
    if (cmodes != 0) {
        while ((cmodes & 1) == 0) {
            cmodes >>= 1;
        }
        if (cmodes != 1) {
            fprintf(stderr, "speaker: ** Audio packet specifies multiple mutually "
                            "exclusive compression modes: %08lX.  Dropped. **\n",
                            msg->compression);
            return;
        }
    }

    /* If message is compressed, decompress appropriately.  Here we get
       quite strict enforcing packet size constraints.  Regardless of the
       provenance of the packet, depending on its compression and protocol
       there is a maximum packet size which will never be exceeded by any
       valid message, and these sizes are well below the threshold at which
       a buffer overflow might occur.  Based on the compression mode, we
       check against the maximum and discard the packet if the contents
       are too long to be valid.  For some protocols we can also test that
       the payload is an even number of encoded frames in the specified
       compression mode.  Notwithstanding these checks, the CODECs still must
       be invulnerable to garbage data which passes these length tests.
       
       Maximum packet payload lengths, by compression type and protocol,
       are as follows.  Protocols with a fixed frame size give the frame
       size after the protocol name.  Many compression types embed the
       actual decompressed length in the first two bytes of the payload,
       accounting for the total being two bytes greater than an even number
       of frames.  ADPCM appends its 3 byte state to the end of the compressed
       audio stream.

                            Speak Freely        RTP         VAT
                            
        Uncompressed            488             320         320
       
        Simple                  488
       
        GSM     [33]            332             134         134
            +Simple             332
         
        ADPCM                   489             323         323
            +Simple             490
        
        LPC     [14]            142              58          58
            +Simple             142
            
        LPC10    [7]             70 
            +Simple              70
            
        CELP    [18]            146
            +Simple             146
            
    */

#define Checklen(n, t)  if (msg->buffer.buffer_len > (n)) { \
            fprintf(stderr, "speaker: ** Payload (%ld) bytes exceeds maximum of %d bytes for " \
                            t " compression. **\n", msg->buffer.buffer_len, n); \
            return; }

#define Checkframe(f, p, t)    if (((msg->buffer.buffer_len - (p)) % (f)) != 0) { \
            fprintf(stderr, "speaker: ** Payload (%ld+%d pad) bytes is not an even number of %d byte frames for " \
                            t " compression. **\n", msg->buffer.buffer_len - (p), p, f); \
            return; }

//    dbg("playbuffer: compression=%x\n", msg->compression);
    if (msg->compression & fCompGSM) {
        Checklen(332, "GSM");
        Checkframe(33, 2, "GSM");
        gsmdecomp(msg);
        len = msg->buffer.buffer_len;
    }

    if (msg->compression & fCompILBC) {
       // Checklen(332, "ILBC");
      //  Checkframe(32, 2, "ILBC");
        ilbcdecomp(msg);
        len = msg->buffer.buffer_len;
    }

    
    /* Update the dynamic sample arrival rate and the exponentially
       smoothed moving average of the same.  To avoid short-term jitter,
       we first average the arrival rate by computing it only after a
       total of DYNRATE_UPDATE_SAMPLE_COUNT samples have arrived (in
       however many packets it takes).  These individual measures
       are then smoothed with an exponentially smoothed moving
       average with smoothing factor DYNRATE_SMOOTHING_FACTOR to
       determine the average arrival rate. */
       
    lastPacketLength += len;

    /* If the jitter idle time has elapsed, reset the jitter
       wait to queue the start of this packet sequence. */

    if (Cjitter > 0 && jitteridle == 0 && jitterwait == 0) {
        jitterwait = Cjitter;
        dbg("speaker: starting %ld millisecond jitter delay.\n", jitterwait / 1000L);
    }

    /* Receipt of any sound buffer resets the jitter idle timer. */

    jitteridle = Cjitteridlet;         /* Reset jitter idle countdown */

    /* If the jitter wait has not yet expired, add the samples to
       the queue waiting to be played at the end of the jitter
       wait.  */

    if (jitterwait > 0) {
        struct jitterbuf *jb;

        jb = (struct jitterbuf *) malloc(sizeof(struct jitterbuf) + len);
        if (jb != NULL) {
            bcopy(val, jb->value, len);
            jb->seq = seq;
            jb->length = len;
            jb->next = NULL;
            if (jittail != NULL) {
                jittail->next = jb;
            }
            jittail = jb;
            if (jithead == NULL) {
                jithead = jb;
                windtimer();
            }
            if (Debug) {
                fprintf(stderr, "speaker: adding %ld samples to jitter queue.\n",
                        len);
            }
#ifdef DEBUG_JITTER            
            dbg("enqueuing buffer %05u delay=%d\n", seq, sound_get_delay());
#endif
        }
    } else {
        flushjitter(0);
        if (audiok) {
            //dbg("playbuffer: len=%d\n", len);
               /* {
                    int i;
                    for (i=0;i<20; i++) dbg("%02x ", (unsigned char) val[i]);
                    dbg("\n");
                }*/
#ifdef DEBUG_JITTER            
            dbg("playing buffer %05u delay=%d\n", seq, sound_get_delay());
#endif
            sound_play(len, (unsigned char *) val);
        }
    }
/*    if (record != NULL) {
        fwrite(val, len, 1, record);
    }*/
}


/*  Main program.  */

void dump_packet(struct soundbuf *sb){

        int compr;
        compr = sb->compression;
        dbg("compression: %08lx ", compr);
        if (compr & fDebug)         dbg("debug ");
        if (compr & fCompGSM)       dbg("gsm ");
        if (compr & fCompILBC)      dbg("ilbc ");
        dbg("\n");
}

int speaker_set(int flags, int ajitter, int ajitteridlet){
    
    if (flags & FLG_DEBUG) debugforce = TRUE;
    if (flags & FLG_HEXDUMP) {
        debugforce = TRUE;
        hexdump = TRUE;
    }

    Cjitter = ajitter;
    Cjitteridlet = (ajitteridlet == 0) ? (Cjitter * 2) : ajitteridlet;
    if (Cjitter < 10000) {
        fprintf(stderr, "speaker: invalid jitter wait %d--reset to 100 milliseconds.\n", Cjitter / 1000);
        Cjitter = 10000;
    }
    if (Cjitteridlet < 10000) {
        fprintf(stderr, "speaker: invalid jitter idle time %d--reset to 100 milliseconds.\n", Cjitteridlet / 1000);
        Cjitteridlet = 10000;
    }
//    dbg("initial jitter=%d\n", jitter);

    return 0;
}

static int addest(const char *line)
{
    struct destination *d;
    zchar *zs;
    char *cmd, *arg;
    
  //  dbg("mike: addest: %s\n", line);

    zs = zstrdup(line);
    cmd = ztokenize(zs, 1);

    d = (struct destination *) malloc(sizeof(struct destination));
	memset(d, 0, sizeof (struct destination));

	arg = ztokenize(zs, 0);
    d->sin.sin_family = AF_INET;
    inet_aton(arg, &d->sin.sin_addr);
	arg = ztokenize(zs, 0);
    //d->sin.sin_port=htons(atoi(arg));
    d->sin.sin_port=0;

	arg = ztokenize(zs, 0);
    d->tcpsock = atoi(arg);
    
	arg = ztokenize(zs, 0);
    d->ssrc = strtoul(arg, NULL, 16);
    
    zfree(zs);
    d->dnext = dests;
    dests = d;
    d->sock = sock;

    return TRUE;
}

static int rmdest(const char *line)
{
    struct destination *d, *prev;
    zchar *zs;
    char *cmd, *arg;
    int tcpsock;
    
   // dbg("mike: rmdest: %s\n", line);

    zs = zstrdup(line);
    cmd = ztokenize(zs, 1);

	arg = ztokenize(zs, 0);
    tcpsock = atoi(arg);
    zfree(zs);

    for (d = prev = dests; d != NULL; prev = d, d = d->dnext) {
        if (d->tcpsock != tcpsock) continue;
        dbg("found socket %d\n", tcpsock);
        if (d==dests) 
            dests = d->dnext;
        else
            prev->dnext = d->dnext;
        free(d);
    }
    return TRUE;
}

void speaker_rel_read(const char *line){
    zchar *zs;
    char *cmd/*, *arg1, *c*/;

    if (strlen(line)==0) {
        error("rel_read: got empty line\n");
        return;
    }

    zs = zstrdup(line);
    cmd = ztokenize(zs, 1);

    if (strcmp(cmd, "q")==0){   
        speaker_running=0;
        goto x;
    }
    if (strcmp(cmd, "0")==0){   
        dbg("speaker stop\n");
        freeaudio();
        goto x;
    }
    if (strcmp(cmd, "1")==0){   
        dbg("speaker start\n");
        //openaudio();
        goto x;
    }
    if (strcmp(cmd, "+")==0){
        //dbg("cmd +\n");
        addest(line);
        goto x;
    }
    if (strcmp(cmd, "-")==0){
        //dbg("cmd -\n");
        rmdest(line);
        goto x;
    }
    
x:;
    zfree(zs);
}

int speaker_run()
{
    struct soundbuf sb;
    char s[1024], *cc;
    char rdbuf[MAXRDBUF];
    int ret, len, rdbuflen;
    struct destination *d;
/*    int i;*/

//    dbg("speker_run: sizeof(soundbuf)=%d r=%d\n", sizeof(struct soundbuf), SIZEOFSOUNDBUF(sb));
    /* Save the number of full seconds at progam start time.  We subtract this
       from packet arrival times measured in microseconds to minimise floating
       point round-off in subsequent computations. */

    if (speaker_pid) {
        dbg("speaker already running\n");
        return 1;
    }
        
    if (pipe(speaker_pipe)){
        error("Can't create speaker pipe\n");
        return -1;
    }
    if (pipe(speaker_rpipe)){
        error("Can't create speaker rpipe\n");
        return -1;
    }
    dbg("speaker_rpipe: %d->%d\n", speaker_rpipe[1], speaker_rpipe[0]);
    speaker_pid=fork();
    if (speaker_pid<0){  /* error */
        error("Can't fork speaker\n");
        return -1;
    }
    if (speaker_pid>0) { /* parent */
        close(speaker_pipe[0]);
        close(speaker_rpipe[1]);
        //dbg("set_handlers(%d) set\n", speaker_rpipe[0]);
        set_handlers(speaker_rpipe[0], speaker_rread, NULL, NULL, NULL);
        return 0;
    }

    /* child */
    close(speaker_pipe[1]);
    close(speaker_rpipe[0]);
    close(0);
    if (dup(speaker_pipe[0])){
        error("Can't dup\n");
        return -1;
    }
    close(speaker_pipe[0]);

    arg0("iprig speaker");
    dbg("speaker started\n");
    
    /*  See if the user has specified a face file.  If so,
        try to open it.  */

    signal(SIGHUP, exiting);          /* Set signal to handle termination */
    signal(SIGINT, exiting);          /* Set signal to handle termination */
    signal(SIGTERM, exiting);         /* Set signal to handle termination */
    signal(SIGALRM, sigalrm);         /* Set signal to handle timeout */
    windtimer();                      /* Set periodic status update  */
        
    gsmh = gsm_create();
    initDecode(&Dec_Inst, 20, 1);


    /* Read from the socket. */
    rdbuflen=0;
    speaker_running=1;

    while (speaker_running) {
        fd_set fdset;
#ifdef USE_RTCP        
        int control;
#endif        
        unsigned long int ssrc;
        u_int16 seq;

        FD_ZERO(&fdset);
        FD_SET(0, &fdset);
        FD_SET(sock, &fdset);
        
        /*  Wait here for a packet to arrive on either the control or
            audio socket.  */
            
        if (select(sock+1, &fdset, NULL, NULL, NULL) <= 0) {
            continue;
        }

        if (FD_ISSET(0, &fdset)){
            ret = read(0, s, sizeof(s));
            if (ret<0) {
                error("speaker: read(0) failed\n");/* maybe SIGPIPE */
                exiting(0);
            }
            if (ret==0) {
                error("speaker: read(0) return 0\n");/* maybe SIGPIPE */
                exiting(0);
            }
            if (rdbuflen + ret >= MAXRDBUF){
                error("speaker: rdbuf overflow\n");
                exiting(0);
            }
            memcpy(rdbuf + rdbuflen, s, ret);
            rdbuflen += ret;

            while(1){
                cc = index(rdbuf, '\n');
                if (!cc) break;
                len = cc - rdbuf;
                rdbuf[len] = '\0';
                if (len>0) {
                    if (rdbuf[len - 1] == '\r') rdbuf[len - 1] = '\0';
                    speaker_rel_read(rdbuf);
                }
                memmove(rdbuf, rdbuf + len + 1, rdbuflen - len - 1);
                rdbuflen -= len + 1;
                
            }
        }

        if (!FD_ISSET(sock, &fdset)) continue;
        
        /*  A packet appears to have arrived.  Read it from the proper socket.  */
        
        fromlen = sizeof(from);
//        dbg("speaker: recvfrom\n");
        if ((rll = recvfrom(sock, (char *) &sb, sizeof sb, 0, (struct sockaddr *) &from, &fromlen)) < 0) {
            if (errno != EINTR) {
                perror("speaker: receiving packet");
            }
            continue;
        }
#ifdef USE_RTCP        
        control = ((unsigned char*)&sb)[1] >= RTCP_SR;
#endif        
        //dbg("control=%d [1]=%d\n", control, ((unsigned char*)&sb)[1]);
#ifdef HEXDUMP
        if (hexdump) {
            fprintf(stderr, "speaker: %d bytes read \n", rll);
            xd(stderr, &sb, rll, TRUE);
        }
#endif      
        if (rll<16) continue;
#ifdef USE_RTCP
        if (control)
            ssrc=get_rtcp_ssrc((unsigned char *)&sb);
        else
#endif     
        {    
            ssrc=get_rtp_ssrc((unsigned char *)&sb);
            seq=get_rtp_seq((unsigned char *)&sb);
        }

//        dbg("recvfrom %s:%d paytype=%3d ssrc=%08x seq=%d\n", inet_ntoa(from.sin_addr), ntohs(from.sin_port), ((unsigned char*)&sb)[1], ssrc, seq);  
        
        //goto found; //FIXME
        for (d = dests; d != NULL; d = d->dnext) {
            if (memcmp(&d->sin.sin_addr, &from.sin_addr, sizeof(struct in_addr)) !=0 ) continue;
//            printf("d->sin.sin_port=%d ssrc=%08lx d->ssrc=%08lx\n", d->sin.sin_port, ssrc, d->ssrc);
            if (d->sin.sin_port == 0){
                char s[50];
                // first packet, remember sin
                if (ssrc != d->ssrc) continue;    
                dbg("remembering %s:%d (old %s:%d)\n", inet_ntoa(from.sin_addr), ntohs(from.sin_port), inet_ntoa(d->sin.sin_addr), ntohs(d->sin.sin_port));  
                memcpy(&d->sin, &from, sizeof (struct sockaddr_in));
                if (opt_s){  // client doesn't update address
                    // send it to main
                    sprintf(s, "PORT;%d;%d\n", d->tcpsock, ntohs(from.sin_port));
                    speaker_rsend(s);
                }
                goto found;
            }else{
                if (memcmp(&d->sin, &from, sizeof(struct sockaddr_in))!=0){
                    error("same ssrc from %s:%d\n", inet_ntoa(d->sin.sin_addr), ntohs(d->sin.sin_port));  
                    break;
                }
                goto found;
            }
        }
//        dbg("not found\n");
        continue;        
found:;
        /* d is valid */
            
#ifdef USE_RTCP
        /* If this packet arrived on the session control port, dispatch
           it to the appropriate handler for its protocol. */

        if (control) {
            unsigned char *p = (unsigned char *) &sb;
            unsigned char *apkt;
            int proto = (p[0] >> 6) & 3;

            if (proto != RTP_VERSION) continue;

            if (isValidRTCPpacket((unsigned char *) &sb, rll)) {

                /* If it's a BYE packet, reset protocol to unknown. */

                if (isRTCPByepacket((unsigned char *) &sb, rll)) {
                    dbg("RTCP BYE %08x\n", get_rtcp_ssrc((unsigned char *)&sb));

                /* If it's a text chat message, print it. */

                } else if (isRTCPAPPpacket((unsigned char *) &sb, rll,
                                RTCP_APP_TEXT_CHAT, &apkt) && apkt != NULL) {

                    dbg("RTCP APP\n");

                /* Otherwise, it's presumably an SDES, from which we
                   should update the user identity information for the
                   connection. */

                } else {
                    //dbg("RTCP SDES %08x\n", get_rtcp_ssrc((unsigned char *)&sb));
                    dbg("RTCP SDES\n");
                } 
            } else {
                dbg("Invalid RTCP\n");
            }
            continue;
        } /* control */
#endif
        if (!isrtp((unsigned char *) &sb, rll)) continue;
        if (sb.buffer.buffer_len == 0) continue;

        playbuffer(&sb, seq);
    }
#ifdef MEANS_OF_EXIT
    close(sock);
    gsm_destroy(gsmh);
    exiting();
/*    if (record != NULL) {
        fclose(record);
    }*/
#endif
    return 0;
}

int speaker_kill(void){
    if (!speaker_pid) return 0;
    set_handlers(speaker_rpipe[0], NULL, NULL, NULL, NULL);

    speaker_send("q\n");
    kill(speaker_pid, SIGINT);

    dbg("waiting for speaker pid %d\n", speaker_pid);
    waitpid(speaker_pid, NULL, 0);
    //dbg("speaker pid %d terminated\n", speaker_pid);
    speaker_pid = 0;
    return(0);
}

int speaker_start(void){
    if (!speaker_pid) {
        error("speaker_start: speaker_pid==0\n");
        return 0;
    }

    speaker_send("1\n");
    return(0);
}

int speaker_stop(void){
    if (!speaker_pid) {
        error("speaker_stop: speaker_pid==0\n");
        return 0;
    }

    speaker_send("0\n");
    return(0);
}

int speaker_addest(struct conn *conn){
    char *c;
    char s1[20], s2[20], s3[20], s4[20];

    strcpy(s1, inet_ntoa(conn->udpsin.sin_addr));
    sprintf(s2, "%d", ntohs(conn->udpsin.sin_port));
    sprintf(s3, "%d", conn->sock); 
    sprintf(s4, "0x%08lx", conn->ssrc); 
    c=zconcatesc("+", s1, s2, s3, s4, NULL);
    speaker_send(c);
    dbg("speaker_addest(%s)\n", c);
    speaker_send("\n");
    mem_free(c);
    return 0;
}

int speaker_rmdest(struct conn *conn){
    char *c;
    char s1[20];

    sprintf(s1, "%d", conn->sock); 
    c=zconcatesc("-", s1, NULL);
    speaker_send(c);
    dbg("speaker_rmdest(%s)\n", c);
    speaker_send("\n");
    mem_free(c);
    return 0;
}

int speaker_getpid(){
    return speaker_pid;
}

/* main process */
void speaker_rread(void *xxx){
    char s[1024], *cc;
    static char rdbuf[MAXRDBUF];
    int ret, len;
    static int rdbuflen = 0;

    ret = read(speaker_rpipe[0], s, sizeof(s));
    dbg("read(%d) %d %d\n", speaker_rpipe[0], ret, errno);
    if (ret<0) {
        error("speaker: rread(%d) failed\n", speaker_rpipe[0]);/* maybe SIGPIPE */
        exiting(0);
    }
    if (ret==0) {
        error("speaker: rread(%d) return 0\n", speaker_rpipe[0]);/* maybe SIGPIPE */
        exiting(0);
    }
    if (rdbuflen + ret >= MAXRDBUF){
        error("speaker: rrdbuf overflow\n");
        exiting(0);
    }
    memcpy(rdbuf + rdbuflen, s, ret);
    rdbuflen += ret;

    while(1){
        cc = index(rdbuf, '\n');
        if (!cc) break;
        len = cc - rdbuf;
        rdbuf[len] = '\0';
        if (len>0) {
            if (rdbuf[len - 1] == '\r') rdbuf[len - 1] = '\0';
            speaker_rrel_read(rdbuf);
        }
        memmove(rdbuf, rdbuf + len + 1, rdbuflen - len - 1);
        rdbuflen -= len + 1;
    }
}

/* main process */
void speaker_rrel_read(const char *line){
    zchar *zs;
    char *cmd, *arg1, *arg2/*, *c */;

    dbg("speaker_rrel_read: '%s'\n", line); 
    if (strlen(line)==0) {
        error("rel_read: got empty line\n");
        return;
    }

    zs = zstrdup(line);
    cmd = ztokenize(zs, 1);

    if (strcmp(cmd, "PORT")==0){
        struct conn *cc;
        int tcpsock, udpport;

        arg1 = ztokenize(zs, 0);
        if (!arg1) goto x;
        arg2 = ztokenize(zs, 0);
        if (!arg2) goto x;

        dbg("cmd PORT\n");
        tcpsock = atoi(arg1);
        udpport = atoi(arg2);

        if (opt_s){
            foreach(cc, conns){
                dbg("  sock=%d\n", cc->sock);
                if (cc->sock != tcpsock) continue;
                dbg("S changing sock=%d %d->%d\n", tcpsock, ntohs(cc->udpsin.sin_port), udpport);
                cc->udpsin.sin_port = htons(udpport);
            }
        }else{
            error("client received PORT command\n");
            /*dbg("  sock=%d\n", sconn->sock);
            if (sconn->sock == tcpsock){
                dbg("C changing sock=%d %d->%d\n", tcpsock, ntohs(sconn->udpsin.sin_port), udpport);
                sconn->udpsin.sin_port = htons(udpport);
            } */
        }
        goto x;
    }
    
x:;
    zfree(zs);
}
