/*

	    Speak Freely for Unix definitions

*/

#include "../config.h"

#define HEXDUMP
#define Internet_Port 2074
#define AUDIO_BLOCKING
#define LINUX
#define NEEDED_LINEAR
#undef USE_RTCP
#undef HAVE_DEV_RANDOM

#if defined(sun) && !defined(Solaris)
#define BSD_like
#endif

#ifdef __FreeBSD__
#define BSD_like
#endif

#ifdef sgi
#define _BSD_SIGNALS
#endif

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>
#ifdef BSD_like
#include <sys/timeb.h>
#endif
#ifdef Solaris
#include <sys/systeminfo.h>
#endif
#ifdef sgi
#include <dmedia/audio.h>
#include <limits.h>
#endif
#include <errno.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#define NOLONGLONG
#include "types.h"



#include "ulaw2linear.h"

#include "../gsm/gsm.h"
#include "../md5/md5.h"
#include "rtp.h"
#include "rtpacket.h"
#include "audio_descr.h"
#include "../ilbc/iLBC_encode.h"
#include "../ilbc/iLBC_decode.h"

#define SAMPLE_RATE 	8000	      /* Audio sample rate in sound packets */

#define BUFL	8000

#define LONG	long

/* Binary block and signal functions for fanatic ex-BSD platforms. */

#ifdef Solaris
#define bcmp(a, b, n)	memcmp(a, b, n)
#define bcopy(a, b, n)	memmove(b, a, n)
#define bzero(a, n)	memset(a, 0, n)

typedef void (*signalFUNC)(int);
#define signal(a, b)	sigset(a, (signalFUNC) b)
#define signalFUNCreturn (signalFUNC)
#endif

#ifndef signalFUNCreturn
#define signalFUNCreturn
#endif


struct soundbuf {
	LONG compression;
	char sendinghost[16];
	struct {
		LONG buffer_len;
		char buffer_val[BUFL];
	} buffer;
};
typedef struct soundbuf soundbuf;

struct sbhead {
	LONG compression;
	char sendinghost[16];
	struct {
		LONG buffer_len;
		char buffer_val[16];
	} buffer;
};
typedef struct sbhead sbhead;

#ifndef TRUE
#define TRUE	1
#endif
#ifndef FALSE
#define FALSE	0
#endif

#define V	(void)

#define ELEMENTS(array) (sizeof(array)/sizeof((array)[0]))

/* Packet mode flags. */

#define fCompILBC   1             /* ILBC compression */
#define fDebug	    2		      /* Debug mode */
#define fKeepAlive  4		      /* Keep alive packet */
#define fCompGSM    32		      /* GSM compression */
#define fEncAlaw    64            /* A-law encoding, 0=mu-law */


/* Mask to extract packet modes. */

#define fCompressionModes   (fCompGSM | fCompILBC )

/* Special definitions for half-duplex control packets. */

#define fHalfDuplex 0x10000000	      /* Local half-duplex transition */

/* Test if a packet actually contains sound. */

#define isSoundPacket(c)    (((c) & (fKeyPGP | fHalfDuplex)) == 0)


/* RTCP APP packet identifiers. */

//#define RTCP_APP_TEXT_CHAT  "SFtc"    /* Text chat message */

/* Look Who's Listening parameters. */

#define LWL_MAX_SITES	    5	      /* Maximum LWL servers a user can publish on */

#ifdef MYDOMAIN
#define getdomainname(s, n)	strncpy(s, MYDOMAIN, n)
#endif


/*  From G711.C  */

#ifdef NEEDED_LINEAR
extern unsigned char linear2alaw(int pcm_val);
extern int alaw2linear(unsigned char a_val);
extern unsigned char linear2ulaw(int pcm_val);
extern int ulaw2linear(unsigned char u_val);
extern unsigned char ulaw2alaw(unsigned char uval);
#endif
extern unsigned char alaw2ulaw(unsigned char aval);


/*  From RTPACKET.C  */

#ifdef RTCP
extern int isValidRTCPpacket(unsigned char *p, int len);
extern int isRTCPByepacket(unsigned char *p, int len);
extern int isRTCPAPPpacket(unsigned char *p, int len, char *name, unsigned char **app_ptr);
extern int rtp_make_bye(unsigned char *p, unsigned long ssrc_i, char *raison, int strict);
extern int rtp_make_app(unsigned char *p, unsigned long ssrc_i, int strict, char *type, char *content);
extern int parseSDES(unsigned char *packet, struct rtcp_sdes_request *r);
extern void copySDESitem(char *s, char *d);
long unsigned int get_rtcp_ssrc(unsigned char *packet);
#endif

extern int isrtp(unsigned char *pkt, int len);
extern int rtpout(soundbuf *sb, unsigned long ssrc_i, unsigned long timestamp_i,
    	    	  unsigned short seq_i, int spurt);
extern int rtp_make_sdes(char **pkt, unsigned long ssrc_i, int port, int strict);
long unsigned int get_rtp_ssrc(unsigned char *packet);
u_int16 get_rtp_seq(unsigned char *packet);

/*  From XDSUB.C  */

#ifdef HEXDUMP
extern void xd(FILE *out, void *buf, int bufl, int dochar);
#endif


/* mike.c */
#define FLG_DEBUG 1
#define FLG_AGC 2
#define FLG_LOOPBACK 4
#define FLG_HEXDUMP 32

/*  Destination host descriptor.  */

struct destination {
    int sock;                         /* Socket file descriptor */
    struct destination *dnext;        /* Next destination in list */
    struct sockaddr_in sin;           /* Internet address */
    int tcpsock;                      /* TCP socket descriptor in parent process */
	unsigned long int ssrc;

};
