/*
    iprig - ham radio via internet
    Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "speakfree.h"
#include "header.h"

#define PROT_VER "1"

int tcpsock;
struct conn conns = {&conns, &conns};
struct conn *sconn = NULL;
struct conn *gconn = NULL;

int init_tcp(){
	int on;
	char errbuf[1024];
	struct sockaddr_in sin;
    
    tcpsock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    /*dbg("   tcpsock=%d\n", net->tcpsock);*/
	if (tcpsock < 0){
        error("Can't create socket for tcp - %s\n", STRERR(errno));
		return -1;
	}

#ifndef __CYGWIN__    
    on=1;
    if (setsockopt(tcpsock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))){
        error("Can't set SO_REUSEADDR - %s\n", STRERR(errno));
        goto x;
    }
#endif    
    
    if (fcntl(tcpsock, F_SETFL, O_NONBLOCK)){
        error("Can't set O_NONBLOCK - %s\n", STRERR(errno));
        goto x;
    }
        
	memset(&sin, 0, sizeof(struct sockaddr_in));
	sin.sin_family = AF_INET;
	sin.sin_port = htons(TCP_PORT);
	sin.sin_addr.s_addr = INADDR_ANY;
	if (bind(tcpsock, (struct sockaddr *)&sin, sizeof(sin))){
		error("Can't bind TCP port %d - %s\n", TCP_PORT, STRERR(errno)); 
		goto x;
	}

    if (listen(tcpsock, 10)){
        error("Can't listen on socket %d, tcp port %d - %s\n", tcpsock, TCP_PORT, STRERR(errno));
        goto x;
    }

    set_handlers(tcpsock, tcp_accept_handler, NULL, tcp_exception_handler, NULL);

    
    return 0;
x:;
    close(tcpsock);
    return -1;
}

void tcp_accept_handler(void *xxx){
    int sock;
    struct sockaddr_in sin;
    socklen_t socklen;
    char *c;
    char salt[100];
    
    struct conn *conn;
    
    
    /*dbg("tcp_accept_handler()\n");*/

    socklen = sizeof(sin);
    sock = accept(tcpsock, (struct sockaddr *)&sin, &socklen);
    if (!socklen || sock<0) return;

    dbg("s%d: accepted %s:%d\n", sock, inet_ntoa(sin.sin_addr), ntohs(sin.sin_port));

    conn = mem_alloc(sizeof(struct conn));
	bzero(conn, sizeof(struct conn));
    
    conn->sin.sin_family      = AF_INET;
    conn->sin.sin_addr.s_addr = sin.sin_addr.s_addr;
    conn->sin.sin_port        = sin.sin_port;

    memcpy(&conn->udpsin, &conn->sin, sizeof(conn->udpsin));
    conn->udpsin.sin_port = htons(Internet_Port);
    
    conn->sock = sock;
    tcp_set_state(conn, TCP_CONNECTED);
    
    set_handlers(conn->sock, tcp_read_handler, 
            NULL, tcp_exception_handler, conn);

	add_to_list(conns, conn);
    
    /* only server */
    c = zconcatesc("VER", PROT_VER, NULL);
    rel_write(conn, c);
    free(c);
    
    sprintf(salt, "%04x%04x%04x%04x", zrandom(), zrandom(), zrandom(), zrandom());
    dbg("SALT='%d'\n", salt);
    c = zconcatesc("SALT", salt, NULL);
    rel_write(conn, c);
    free(c);
    conn->salt = strdup(salt);
}

void client_authorized(struct conn *conn){
    char s[25];
    char *c;

    conn->authorized = 1;
    sprintf(s, "0x%08lx", mike_getssrc());
    c = zconcatesc("SSRC", s, NULL);
    rel_write(conn, c);
    free(c);
    rel_write(conn, "AUTHOK");

    speaker_addest(conn);
    mike_addest(conn);
	
	if (!gconn) {
		dbg("grabbing control\n");
		gconn = conn;
		speaker_stop();
		mike_start();
        send_grab(conn, 1);
	}else{
		dbg("only rx mode\n");
		if (rig->ptt){
			/* todo restart speaker to take effect*/
		}
	}

    tcp_send_state(conn);
}

void tcp_connect(char *hostname, int port){
    int ret;
	char errbuf[1024];
    struct conn *conn;
    struct hostent *he;
        

    he = gethostbyname(hostname);
    if (!he){
        error("Can't resolve %s\n", hostname);
        return;
    }
    
    sconn = conn = mem_alloc(sizeof(struct conn));
	bzero(conn, sizeof(struct conn));

    conn->sin.sin_family      = AF_INET;
    memcpy(&conn->sin.sin_addr, he->h_addr_list[0], he->h_length);
    conn->sin.sin_port        = htons(port);
    
    memcpy(&conn->udpsin, &conn->sin, sizeof(conn->udpsin));
    conn->udpsin.sin_port = htons(Internet_Port);
    
    conn->prev = NULL;
    conn->next = NULL;
    
/*    dbg("tcp->connect %s:%d\n", inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));  */
   
    conn->sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
/*    dbg("socket returns %d \n", conn->sock);*/
    
    if (fcntl(conn->sock, F_SETFL, O_NONBLOCK)){
        error("Can't set O_NONBLOCK\n");
        return;
    }
    
    ret = connect(conn->sock, (struct sockaddr *) &conn->sin, sizeof(struct sockaddr_in));
/*    dbg("connect returns %d (%d) '%s'\n", ret, errno, strerror_r(errno));*/
    if (ret==0){
        set_handlers(conn->sock, tcp_read_handler, NULL, tcp_exception_handler, conn);
        tcp_set_state(conn, TCP_CONNECTED);
        after_connect(conn);
        return;
    }
    if (ret<0){
        if (errno==EALREADY || errno==EINPROGRESS){
            tcp_set_state(conn, TCP_CONNECTING);
            set_handlers(conn->sock, NULL, tcp_connected_handler, tcp_exception_handler, conn);
            dbg("s%d: connecting to %s:%d\n", conn->sock, inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));
            return;
        }  
        error("connect error %s\n", STRERR(errno));
        tcp_disconnect(conn, "connect error");
    }
}

void tcp_connected_handler(void *xxx){
    struct conn *conn;
    int ret, err, cnt;
    socklen_t optlen;
    char errbuf[1024], s[1300];
    
    
    conn = (struct conn *)xxx;
/*    dbg("tcp_connected_handler %s:%d\n", inet_ntoa(conn->sin.sin_addr),
 *    ntohs(conn->sin.sin_port));*/
    
    if (fcntl(conn->sock, F_SETFL, 0)){
        tcp_disconnect(conn, "Can't clear O_NONBLOCK");
        return;
    }

    cnt=10;
gsopt:;    
    optlen = sizeof(err);
    ret = getsockopt(conn->sock, SOL_SOCKET, SO_ERROR, (void *)&err, &optlen);
/*    dbg("   getsockopt(SO_ERROR) returns %d err is %d\n", ret, err);  */
    if (ret!=0){
        tcp_disconnect(conn, "getsockopt returns !=0");
        return;
    }
    if (err!=0){
        sprintf(s, "getsockopt returns error %d - %s", err, STRERR(errno));
        if (errno==EINPROGRESS && --cnt>0) goto gsopt;
        tcp_disconnect(conn, s);
        return;
    }
    
    if (fcntl(conn->sock, F_SETFL, O_NONBLOCK)){
        tcp_disconnect(conn, "Can't set O_NONBLOCK");
        return;
    }
    tcp_set_state(conn, TCP_CONNECTED);
    set_handlers(conn->sock, tcp_read_handler, 
            NULL, tcp_exception_handler, conn);
    after_connect(conn);
}

void after_connect(struct conn *conn){
    char *c;
    dbg("s%d: connected to %s:%d\n", conn->sock, inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));
    
    /* only client */
    c = zconcatesc("VER", PROT_VER, NULL);
    rel_write(conn, c);
    free(c);
    
}

void tcp_disconnect(struct conn *conn, char *reason){
    struct conn *cc;

    if (conn->disconnecting) return;
    conn->disconnecting=1;

	dbg("s%d: disconnected %s:%d because of %s\n", conn->sock, inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port), reason);
    set_handlers(conn->sock, NULL, NULL, NULL, NULL);
    if (conn->timer) kill_timer(conn->timer);
    if (conn == sconn) sconn = NULL;
    if (conn == gconn) gconn = NULL;
    if (conn->prev) del_from_list(conn);

    if (opt_s){
        mike_rmdest(conn);
        if (!gconn){
            foreach(cc, conns) gconn=cc;
            if (gconn){
                send_grab(gconn, 1);
	            dbg("new grab %d %s:%d\n", gconn->sock, inet_ntoa(gconn->sin.sin_addr), ntohs(gconn->sin.sin_port));
            }
        }
    }else{
        mike_rmdest(conn);
        terminate=1;
    }
    
    close(conn->sock);
	mem_free(conn);
}

void tcp_read_handler(void *xxx){
    struct conn *conn;            
    char s[1030], *d;
    int ret, err, len;
    char errbuf[1030];

    conn = (struct conn*)xxx;
    ret=read(conn->sock, s, 1024);
    err=errno;
	if (ret<0){
		dbg("s%d: error reading socket - %s\n", conn->sock, STRERR(errno));
		tcp_disconnect(conn, "read returns <0");
		return;
	}
	if (ret==0){
		dbg("s%d: got disconnect\n", conn->sock);
		tcp_disconnect(conn, "got disconnect");
		return;
	}
	
	if (conn->rdbuflen + ret >= MAXRDBUF){
        error("socket %d rdbuf overflow (%d>%d)\n", conn->sock, conn->rdbuflen+ret, MAXWRBUF);
		tcp_disconnect(conn, "rdbuf overflow");
		return;
    }
	memcpy(conn->rdbuf + conn->rdbuflen, s, ret);
	conn->rdbuflen += ret;

    if (opt_d>=3) {
        char s[MAXRDBUF+1];
        int i;
        
        memcpy(s, conn->rdbuf, conn->rdbuflen);
        for (i=0;i<conn->rdbuflen;i++) {
            if (!isprint(s[i])) s[i]='.';
        }
        s[i]='\0';
        
        dbg("s%d: read '%s'\n", conn->sock, s );
        /*dbg("s%d: rel_write '%s' (%d/%d/%d)", conn->sock, conn->wrbuf, conn->wrbuflen, written,err);*/
    }
	
    while(1){
        d = index(conn->rdbuf, '\n');
        if (!d) break;
        len = d - conn->rdbuf;
        conn->rdbuf[len] = '\0';
        if (len>0) {
            if (conn->rdbuf[len - 1] == '\r') conn->rdbuf[len - 1] = '\0';
            rel_read(conn, conn->rdbuf);
        }
        memmove(conn->rdbuf, conn->rdbuf + len + 1, conn->rdbuflen - len - 1);
        conn->rdbuflen -= len + 1;
        
    }
	return;
}

void tcp_exception_handler(void *xxx){
    struct conn *conn;
    
    conn = (struct conn *)xxx;
    dbg("s%d: tcp_exception_handler %s:%d\n", conn->sock, inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port)); 
    
    tcp_disconnect(conn, "socket exception");
}

int tcp_send_state(struct conn *conn){
	
	return 0;
}

int conn_prod_state(struct conn *conn){
    if (!conn) return 0;
    if (conn->state == TCP_CONNECTED || conn->state == TCP_WAIT_ACK ) return 1;
    return 0;
}

void tcp_set_state(struct conn *conn, enum tcp_state state){
    
    if (opt_d>=3) dbg("s%d: tcp_set_state %s -> %s \n", conn->sock, tcp_state_s(conn->state), tcp_state_s(state));
    
    if (conn->timer) kill_timer(conn->timer);
    conn->timer = 0;
    conn->state = state;
    switch (conn->state){
        case TCP_INIT:
            break;
        case TCP_CONNECTING:
            conn->timer = install_timer(CONNECTION_TIMEOUT, kill_conn_timer, conn);
            break;
        case TCP_CONNECTED:
            conn->timer = install_timer(SEND_ACK_TIMEOUT, send_ack_timer, conn);
            break;
        case TCP_WAIT_ACK:
            conn->timer = install_timer(ACK_TIMEOUT, kill_conn_timer, conn);
            break;
        case TCP_DISCONNECTED:
            conn->timer = install_timer(REMOVE_TIMEOUT, remove_conn_timer, conn);
            break;
        case TCP_DEAD:
            break;    
    }
}

char *tcp_state_s(enum tcp_state state){
    switch (state){
        case TCP_INIT:
            return "INIT";
        case TCP_CONNECTING:
            return "CONNECTING";
        case TCP_CONNECTED:
            return "CONNECTED";
        case TCP_WAIT_ACK:
            return "WAIT_ACK";
        case TCP_DISCONNECTED:
            return "DISCONNECTED";
        case TCP_DEAD:
            return "DEAD";
    }
    return "???";
}

void send_ack_timer(void *xxx){
    struct conn *conn;
    char s[20], *c;

    conn = (struct conn *)xxx;
    if (opt_d>=3) dbg("s%d: send_ack_timer %s:%d\n", conn->sock, inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));
    conn->timer = 0;
    
    snprintf(s, 19, "%d", ++conn->relseq);
    c = zconcatesc("KA", s, NULL);
/*    dbg("s='%s' c='%s'\n", s, c);*/
    rel_write(conn, c);
    mem_free(c);
    udp_keepalive(conn);
    tcp_set_state(conn, TCP_WAIT_ACK);
}

void kill_conn_timer(void *xxx){
    struct conn *conn;

    conn = (struct conn *)xxx;
    dbg("s%d: kill_conn_timer %s:%d\n", conn->sock, inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));
    conn->timer = 0;
    tcp_disconnect(conn, "connection or ack timeout");
}

void remove_conn_timer(void *xxx){
    struct conn *conn;

    conn = (struct conn *)xxx;
/*    dbg("remove_conn_timer %s:%d\n", inet_ntoa(conn->sin.sin_addr),
 *    ntohs(conn->sin.sin_port));*/
    conn->timer = 0;

    del_from_list(conn);
    free_conn(conn);
    mem_free(conn);
}

void free_conn(struct conn *conn){
    dbg("s%d: free_conn\n", conn->sock);
    set_handlers(conn->sock, NULL, NULL, NULL, NULL);
    if (conn->sock>=0) close(conn->sock);
    conn->sock=-1;
    if (conn->salt) free(conn->salt);
    if (conn->user) free(conn->user);
    tcp_set_state(conn, TCP_DEAD);
}

void tcp_write_handler(void *xxx){
    struct conn *conn;
    int written, err;
    char errbuf[1030];

    conn = (struct conn *)xxx;
/*    dbg("tcp_write_handler %s:%d\n", inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));*/
    
    if (!conn->wrbuflen){
        set_handlers(conn->sock, tcp_read_handler, 
                NULL, tcp_exception_handler, conn);
        return;
    }
    written = write(conn->sock, conn->wrbuf, conn->wrbuflen);
    err=errno;
    if (opt_d>=3) dbg("s%d: tcp_write_handler '%s' (%d/%d) %s", conn->sock, conn->wrbuf, conn->wrbuflen, written,err);
    /*dbg("    towrite=%d  written=%d \n", towrite, written); */
    if (written<=0){
        error("error writing to socket %d - %s\n", conn->sock, STRERR(err));
        tcp_disconnect(conn, "write returns <=0");
        return;
    }
    if (conn->wrbuflen == written){
        conn->wrbuflen = 0;
        set_handlers(conn->sock, tcp_read_handler, 
                NULL, tcp_exception_handler, conn);
        return;
        
    }
    memmove(conn->wrbuf + written, conn->wrbuf, conn->wrbuflen - written); 
    conn->wrbuflen -= written;

}

/***************** "RELATION" LAYER ********************************/

void rel_write(struct conn *conn, char *s){
    int written,err, len;
    char errbuf[1030];
    
    if (!conn_prod_state(conn)){
        error("connection not i production state\n");
        return;
    }
    
    len=strlen(s);
    if (conn->wrbuflen + len + 1 > MAXWRBUF){
        error("socket %d wrbuf overflow (%d>%d)", conn->sock, conn->wrbuflen + len + 1, MAXWRBUF);
        tcp_disconnect(conn, "wrbuf overflow");
        return;
    }
    memcpy(conn->wrbuf, s, len);
    conn->wrbuf[len] = '\n';
    conn->wrbuflen += len + 1;

    written = write(conn->sock, conn->wrbuf, conn->wrbuflen);
    err=errno;
    if (opt_d>=3) {
        char s[MAXWRBUF+1];
        int i;
        
        memcpy(s, conn->wrbuf, conn->wrbuflen);
        for (i=0;i<conn->wrbuflen;i++) {
            if (!isprint(s[i])) s[i]='.';
        }
        s[i]='\0';
        
        dbg("s%d: rel_write '%s' (%d/%d/%d)\n", conn->sock, s, conn->wrbuflen, written,err);
        /*dbg("s%d: rel_write '%s' (%d/%d/%d)", conn->sock, conn->wrbuf, conn->wrbuflen, written,err);*/
    }
    if (written<0){
        if (err==EAGAIN){
            set_handlers(conn->sock, tcp_read_handler, 
                    tcp_write_handler, tcp_exception_handler, conn);
            return;
        }
        error("error writing to socket %d - %s\n", conn->sock, STRERR(err));
        tcp_disconnect(conn, "write returns <0, errno!=EAGAIN");
        return;
    }
        
    if (conn->wrbuflen == written){
        conn->wrbuflen=0;
        return;
    }
    memmove(conn->wrbuf + written, conn->wrbuf, conn->wrbuflen - written);
    conn->wrbuflen -= written;
    set_handlers(conn->sock, tcp_read_handler, 
            tcp_write_handler, tcp_exception_handler, conn);
                
}

/*void rel_write_all(char *s){
    int i;
    struct conn *conn;
    char ss[256];
    
    strncpy(ss, s, 2);
    ss[3]='\0';

    
    for (i=0; i<net->peers->len; i++){
        conn = (struct conn *) g_ptr_array_index(net->peers, i);    
        if (conn_prod_state(conn)) {
            dbg("%d(%s) ", conn->sock,tcp_state_s(conn) );
            rel_write(conn, s);
        }
    }
    dbg("\n");
} */

void rel_read(struct conn *conn, char *line){
    zchar *zs;
    char *cmd, *arg1, *c;

    if (strlen(line)==0) {
        error("rel_read: got empty line\n");
        return;
    }

    zs = zstrdup(line);
    cmd = ztokenize(zs, 1);

    if (opt_d>=3) dbg("s%d: rel_read '%s'\n", conn->sock, line);
    
    if (strcmp(cmd, "KA")==0){    /* keepalive */
/*        dbg("   recvd KA, sending AK\n");*/
        arg1 = ztokenize(zs, 0);

        c = zconcatesc("AK", arg1, NULL);
        //dbg("arg1='%s' c='%s'\n", arg1, c);
        rel_write(conn, c);
        free(c);
        goto x;
    }

    if (strcmp(cmd, "AK")==0){   /* keepalive ack */
/*        dbg("   recvd AK, setting NS_CONNECTED\n");*/
        /* todo AK number */
        tcp_set_state(conn, TCP_CONNECTED);
        
        goto x;
    }
    
  /************** app layer ********************/  
    
    if (opt_s) {
        if (opt_d==2) dbg("s%d: rel_sread '%s'\n", conn->sock, line);
        rel_sread(conn, cmd, zs);
    }else{
        if (opt_d==2) dbg("s%d: rel_cread '%s'\n", conn->sock, line);
        rel_cread(conn, cmd, zs);
    }
/*    dbg("app_read sock=%d '%s'\n", conn->sock, line);*/
    
    
x:;
    zfree(zs);
}

void rel_sread(struct conn *conn, char *cmd, zchar *zs){
    char *c, *arg1, *arg2/*, *arg3*/;
    struct MD5Context md5;
    
    if (strcmp(cmd, "AUTH")==0){        /* authentization from client */
        char key[100], *auth1, *auth2;

        arg1 = ztokenize(zs, 0);        /* user */
        if (!arg1 || !*arg1){
            tcp_disconnect(conn, "auth: empty username");
            return;
        }
        if (strlen(arg1)>8) {
            tcp_disconnect(conn, "auth: too long username");
            return;
        }
        arg2 = ztokenize(zs, 0);        /* auth */
        if (!arg2 || strlen(arg2)!=32){
            tcp_disconnect(conn, "auth: bad auth");
            return;
        }
        strcpy(key, "user_");
        strcat(key, arg1);

//      rc_getopt("user_ok1zia"); /* MD5 hash of OK1ZIA's password */        
        c = rc_getopt(key); 
        if (!c || !*c){
            tcp_disconnect(conn, "auth: unknown user");
            return;
        }
        if (!conn->salt) {
            tcp_disconnect(conn, "auth: bad salt");
            return;
        }
        
        auth1 = zconcatesc(arg1, c, conn->salt, NULL);
        MD5Init(&md5);
        dbg("auth1='%s'\n", auth1);
        MD5Update(&md5, (unsigned char *)auth1, strlen(auth1));
        auth2 = (char *)MD5Final_str(&md5);
        dbg("computed auth1='%s' auth2='%s' arg2='%s'\n", auth1, auth2, arg2);
        free(auth1);

        if (strcmp(auth2, arg2)!=0){
            tcp_disconnect(conn, "auth: bad password");
        // FIXME    return;
        }
        conn->user=strdup(arg1);
        client_authorized(conn);
    }
    
    if (strcmp(cmd, "SSRC")==0){        /* synchronization source identifier */
        arg1 = ztokenize(zs, 0);
        conn->ssrc = strtoul(arg1, NULL, 16);
        dbg("sread: recvd SSRC %08lx\n", conn->ssrc);
    }

    if (!conn->authorized) return;

    /*** only authorized connections */


    if (conn!=gconn) return;

    /*** only for grabbed control ***/

    if (strcmp(cmd, "PTT")==0){     /* push to talk request */
		arg1 = ztokenize(zs, 0);
		if (!arg1){
			error("missing argument 1 for PTT\n");
			return;
		}
		
        if (rig->ptt){
			rig->ptt = 0;
            rig_ptt(0);
			speaker_stop();
			mike_start();
			
		}else{
			rig->ptt = 1;
            rig_ptt(1);
			mike_stop();
			speaker_start();
			rel_write(conn, "RMIKE");
		}
        return;
    }
    
    if (strcmp(cmd, "CWS")==0){     /* straight cw key */
		arg1 = ztokenize(zs, 0);
		if (!arg1){
			error("missing argument 1 for CWS\n");
			return;
		}
        rig->cw=atoi(arg1);
        rig_cw(rig->cw);
    }

    if (strcmp(cmd, "PWR")==0){     /* power supply */
		arg1 = ztokenize(zs, 0);
		if (!arg1){
			error("missing argument 1 for PWR\n");
			return;
		}
        rig->pwr=atoi(arg1);
        rig_pwr(rig->pwr);
    }

    if (strcmp(cmd, "LNA")==0){     /* LNA */
		arg1 = ztokenize(zs, 0);
		if (!arg1){
			error("missing argument 1 for LNA\n");
			return;
		}
        rig->lna=atoi(arg1);
        rig_lna(rig->lna);
    }

    if (strcmp(cmd, "PA")==0){     /* PA */
		arg1 = ztokenize(zs, 0);
		if (!arg1){
			error("missing argument 1 for PA\n");
			return;
		}
        rig->pa=atoi(arg1);
        rig_pa(rig->pa);
    }

    if (strcmp(cmd, "GPO")==0){     /* Output 0 */
        int gpo;

		arg1 = ztokenize(zs, 0);
		if (!arg1){
			error("missing argument 1 for GPO\n");
			return;
		}
		arg2 = ztokenize(zs, 0);
		if (!arg1){
			error("missing argument 2 for GPO\n");
			return;
		}
        gpo=atoi(arg1);
        if (gpo<0 || gpo>MAXGPO){
			error("bad gpo %s\n", arg1);
			return;
        }
        rig->gpo[gpo]=atoi(arg2);
        rig_gpo(gpo, rig->gpo[gpo]);
    }

    if (strcmp(cmd, "PWK")==0){     /* power button on panel */
		arg1 = ztokenize(zs, 0);
		if (!arg1){
			error("missing argument 1 for PWK\n");
			return;
		}
        rig->pwk=atoi(arg1);
        rig_pwk(rig->pwk);
    }
}

void rel_cread(struct conn *conn, char *cmd, zchar *zs){
    char *arg1/*, *arg2, *arg3*/;
    char *c, *auth1, *auth2, *user, *pass;
    char s[25];
    struct MD5Context md5;
    
    if (strcmp(cmd, "SALT")==0){        /* salt from server */
        arg1 = ztokenize(zs, 0);

        user = rc_getopt("user");       /* Client user name (callsign) */
        pass = rc_getopt("pass");       /* Client password (MD5 hash) */
        if (!user || !*user || !pass || !*pass){
            error("Unknown user or pass");
            tcp_disconnect(conn, "unknown user or pass");
            return;
        }
        //dbg("user=%s pass=%s\n", user, pass);
        auth1 = zconcatesc(user, pass, arg1, NULL);
        MD5Init(&md5);
        dbg("auth1='%s'\n", auth1);
        MD5Update(&md5, (unsigned char*)auth1, strlen(auth1));
        free(auth1);

        sprintf(s, "0x%08lx", mike_getssrc());
        c = zconcatesc("SSRC", s, NULL);
        rel_write(conn, c);
        dbg("cread: rel_write(%s)\n", c);
        free(c);

        auth2 = zconcatesc("AUTH", user, MD5Final_str(&md5), NULL);
        rel_write(conn, auth2);
        free(auth2);
        return;
    }

    if (strcmp(cmd, "SSRC")==0){        /* synchronization source identifier */
        arg1 = ztokenize(zs, 0);
        conn->ssrc = strtoul(arg1, NULL, 16);
    }
    
    if (strcmp(cmd, "AUTHOK")==0){      /* auth success */
        speaker_addest(conn);
        mike_addest(conn);
        mike_stop();
        speaker_start();
    }

    if (strcmp(cmd, "RMIKE")==0){       /* run mike */
		mike_start();
        return;
    }

    if (strcmp(cmd, "GRAB")==0){        /* controlling of rig on server */
        arg1 = ztokenize(zs, 0);
		if (!arg1){
			error("missing argument 1 for GRAB\n");
			return;
		}
        crig->grab=atoi(arg1);
        crig_grab();
    }
}

void send_grab(struct conn *conn, int grab){
    char s[20], *c;

    snprintf(s, 19, "%d", grab);
    c = zconcatesc("GRAB", s, NULL);
    rel_write(conn, c);
    mem_free(c);
}
